# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# MIT License

# Copyright (c) 2020 Haotian Liu and Rafael A. Rivera-Soto

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""YolactEdge model with the feature pyramid network and prediction layer.
Modifications include:
- Yolact
    - Refactor necessary configs from yolact_edge/data/config.py
    - Removed unused conditions for ResNet and MobileNetV2 backbone
        - use_jit boolean value
    - Removed unused functions for ResNet and MobileNetV2 backbone
        - save_weights()
        - init_weights()
        - train()
        - freeze_bn()
        - fine_tune_layers()
        - extra_loss()
        - forward_flow()
        - create_embed_flow_net()
        - create_partial_backbone()
    - Removed unused functions for TRT implementation
        - _get_trt_cache_path()
        - has_trt_cached_module()
        - load_trt_cached_module()
        - save_trt_cached_module()
        - trt_load_if()
        - to_tensorrt_protonet()
        - to_tensorrt_fpn()
        - to_tensorrt_prediction_head()
        - to_tensorrt_spa()
        - to_tensorrt_flow_net()
- PredictionModule
    - Removed unused make_priors function
- Removed unused Concat class
- Removed unused PredictionModuleTRT class
- Removed unused Cat class
- Removed unused ShuffleCat class
- Removed unused ShuffleCatChunk class
- Removed unused ShuffleCatAlt class
- Removed unused FlowNetUnwrap class
- Removed unused FlowNetMiniTRTWrapper class
- Removed unused PredictionModuleTRTWrapper class
- Removed unused NoReLUBottleneck class
- Removed unused FlowNetMiniPredLayer class
- Removed unused FlowNetMiniPreConvs class
- Removed unused FlowNetMini class
- Removed unused FlowNetMiniTRT class
- Removed unused SPA class
- Removed unused FPN class
- Removed traditional NMS
"""

import logging
from math import sqrt
from itertools import product
from pathlib import Path
from typing import Iterable, List, Tuple, Dict, Any, Union, Optional, Callable

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from peekingduck.pipeline.nodes.model.yolact_edgev1.yolact_edge_files.utils import (
    make_net,
    jaccard,
    decode,
    make_extra,
)
from peekingduck.pipeline.nodes.model.yolact_edgev1.yolact_edge_files.backbone import (
    ResNetBackbone,
    MobileNetV2Backbone,
)

if torch.cuda.is_available():
    torch.cuda.current_device()


ScriptModuleWrapper = torch.jit.ScriptModule

NUM_DOWNSAMPLE = 2
FPN_NUM_FEATURES = 256
NUM_CLASSES = 81  # For COCO classes


class YolactEdge(nn.Module):  # pylint: disable=too-many-instance-attributes
    """YolactEdge model module.

    Values for the selected backbone layers, channels, prediction aspect ratios,
    and scales are taken from the configuration file in the original repository:
    https://github.com/haotian-liu/yolact_edge/blob/master/yolact_edge/data/config.py
    """

    def __init__(
        self,
        model_type: str,
        input_size: int,
        iou_threshold: float,
        max_num_detections: int,
    ) -> None:
        super().__init__()

        self.backbone: Union[ResNetBackbone, MobileNetV2Backbone]

        if model_type[0] == "r":
            if model_type == "r101-fpn":
                self.backbone = ResNetBackbone(([3, 4, 23, 3]))
            elif model_type == "r50-fpn":
                self.backbone = ResNetBackbone(([3, 4, 6, 3]))
            self.layers = list(range(1, 4))
            src_channels = [256, 512, 1024, 2048]
        elif model_type == "mobilenetv2":
            self.backbone = MobileNetV2Backbone(
                1.0,
                [  # Based on MobileNetV2 paper's bottleneck values for t,c,n,s
                    [1, 16, 1, 1],
                    [6, 24, 2, 2],
                    [6, 32, 3, 2],
                    [6, 64, 4, 2],
                    [6, 96, 3, 1],
                    [6, 160, 3, 2],
                    [6, 320, 1, 1],
                ],
                8,
            )
            self.layers = [3, 4, 6]
            src_channels = [32, 16, 24, 32, 64, 96, 160, 320, 1280]

        num_layers = max(self.layers) + 1
        while len(self.backbone.layers) < num_layers:
            self.backbone.add_layer()

        self.num_grids = 0
        self.proto_src = 0
        mask_proto_net = (
            [(256, 3, {"padding": 1})] * 3
            + [(None, -2, {}), (256, 3, {"padding": 1})]  # type: ignore
            + [(32, 1, {})]
        )

        self.proto_net, _ = make_net(256, mask_proto_net, include_last_relu=False)

        self.fpn_phase_1 = FPNPhase1([src_channels[i] for i in self.layers])
        self.fpn_phase_2 = FPNPhase2([src_channels[i] for i in self.layers])

        self.selected_layers = list(range(len(self.layers) + NUM_DOWNSAMPLE))

        # The following values work for ResNet and MobileNetV2 backbones. Other
        # backbones may require different values.
        self.pred_aspect_ratios = [[[1, 1 / 2, 2]]] * 5
        self.pred_scales = [[24], [48], [96], [192], [384]]

        src_channels = [FPN_NUM_FEATURES] * len(self.selected_layers)
        self.prediction_layers = nn.ModuleList()

        for idx, layer_idx in enumerate(self.selected_layers):
            parent = None
            if idx > 0:
                parent = self.prediction_layers[0]
            pred = PredictionModule(
                src_channels[layer_idx],
                src_channels[layer_idx],
                aspect_ratios=self.pred_aspect_ratios[idx],
                scales=self.pred_scales[idx],  # type: ignore
                parent=parent,
                index=idx,
                input_size=input_size,
            )
            self.prediction_layers.append(pred)

        self.semantic_seg_conv = nn.Conv2d(
            src_channels[0], NUM_CLASSES - 1, kernel_size=1
        )
        self.detect = YolactEdgeHead(
            NUM_CLASSES,
            bkg_label=0,
            conf_thresh=0.05,
            iou_threshold=iou_threshold,  # This is the same as nms_thresh
            max_num_detections=max_num_detections,
        )

    def forward(self, inputs: Tensor) -> Dict[str, List]:
        """The input should be of size [batch_size, 3, img_h, img_w]

        Args:
            inputs (Tensor): The input tensor

        Returns:
            outs_wrapper (Dict): Prediction output for YolactEdge
        """
        outs_wrapper = {}
        outs = self.backbone(inputs)
        outs = [outs[i] for i in self.layers]
        outs_fpn_phase_1_wrapper = self.fpn_phase_1(*outs)
        outs_phase_1, _ = (
            outs_fpn_phase_1_wrapper[: len(outs)],
            outs_fpn_phase_1_wrapper[len(outs) :],
        )
        outs_wrapper["outs_phase_1"] = [out.detach() for out in outs_phase_1]
        outs = self.fpn_phase_2(*outs_phase_1)
        outs_wrapper["outs_phase_2"] = [out.detach() for out in outs]

        proto_x = inputs if self.proto_src is None else outs[self.proto_src]
        proto_out = self.proto_net(proto_x)
        proto_out = torch.nn.functional.relu(proto_out, inplace=True)
        proto_out = proto_out.permute(0, 2, 3, 1).contiguous()

        pred_outs: Dict[str, Any]
        pred_outs = {"loc": [], "conf": [], "mask": [], "priors": []}

        for idx, pred_layer in zip(self.selected_layers, self.prediction_layers):
            pred_x = outs[idx]
            for key, val in pred_layer(pred_x).items():
                pred_outs[key].append(val)

        for key, val in pred_outs.items():
            pred_outs[key] = torch.cat(val, -2)

        pred_outs["proto"] = proto_out
        pred_outs["conf"] = F.softmax(pred_outs["conf"], -1)
        outs_wrapper["pred_outs"] = self.detect(pred_outs)
        return outs_wrapper

    def load_weights(self, path: Path) -> None:
        """Loads weights from a compressed save file.

        Args:
            path (Path): Path to the model weights file.

        Returns:
            YolactEdge model
        """
        state_dict = torch.load(path, map_location="cpu")
        for key in list(state_dict.keys()):
            # For backward compatibility, the new variable is called layers.
            # This has been commented out because the ResNet and MobileNetV2
            # backbones will not be using it.
            # if key.startswith("backbone.layer") and not key.startswith(
            #     "backbone.layers"
            # ):
            #     del state_dict[key]
            if key.startswith("fpn.downsample_layers."):
                if int(key.split(".")[2]) >= NUM_DOWNSAMPLE:
                    del state_dict[key]
            if key.startswith("fpn.lat_layers"):
                state_dict[key.replace("fpn.", "fpn_phase_1.")] = state_dict[key]
                del state_dict[key]
            elif key.startswith("fpn.") and key in state_dict:
                state_dict[key.replace("fpn.", "fpn_phase_2.")] = state_dict[key]
                del state_dict[key]
        self.load_state_dict(state_dict)


class PredictionModule(nn.Module):  # pylint: disable=too-many-instance-attributes
    """
    The (c) prediction module adapted from DSSD:
    https://arxiv.org/pdf/1701.06659.pdf
    Note that this is slightly different to the module in the paper
    because the Bottleneck block actually has a 3x3 convolution in
    the middle instead of a 1x1 convolution. Though, I really can't
    be arsed to implement it myself, and, who knows, this might be
    better.

    Args:
        in_channels (int): The input feature size.
        out_channels (int): The output feature size (must be a multiple of 4).
        aspect_ratios (List[List]): A list of lists of priorbox aspect ratios
            (one list per scale).
        scales (List[int]): A list of priorbox scales relative to this layer's convsize.
            For instance: If this layer has convouts of size 30x30 for an image
            of size 600x600, the 'default' (scale of 1) for this layer would
            produce bounding boxes with an area of 20x20px. If the scale is .5
            on the other hand, this layer would consider bounding boxes with area
            10x10px, etc.
        parent (PredictionModule): If parent is a PredictionModule, this module
            will use all the layers from parent instead of from this module.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        in_channels: int,
        out_channels: int = 1024,
        aspect_ratios: Iterable[Any] = None,
        scales: Iterable[List[int]] = None,
        parent: Optional[Callable] = None,  # PredictionModule
        index: int = 0,
        input_size: int = None,
    ) -> None:

        super().__init__()
        self.params = [in_channels, out_channels, aspect_ratios, scales, parent, index]
        self.num_classes = NUM_CLASSES
        self.mask_dim = 32
        self.num_priors = sum(len(x) for x in aspect_ratios)  # type: ignore
        self.parent = [parent]
        self.index = index
        head_layer_params: Dict[Any, Any] = {"kernel_size": 3, "padding": 1}

        if parent is None:
            self.upfeature, out_channels = make_net(
                in_channels, [(256, 3, {"padding": 1})]
            )
            self.bbox_layer = nn.Conv2d(
                out_channels, self.num_priors * 4, **head_layer_params
            )
            self.conf_layer = nn.Conv2d(
                out_channels, self.num_priors * self.num_classes, **head_layer_params
            )
            self.mask_layer = nn.Conv2d(
                out_channels, self.num_priors * self.mask_dim, **head_layer_params
            )
            self.bbox_extra, self.conf_extra, self.mask_extra = [
                make_extra(x, out_channels) for x in (0, 0, 0)
            ]

        self.aspect_ratios = aspect_ratios
        self.scales = scales
        self.priors: Union[None, Tensor] = None
        self.last_conv_size = (0, 0)
        self.input_size = input_size

    def forward(self, inputs: Tensor) -> Dict[str, Tensor]:
        """
        Args:
            inputs (Tensor): The convout from a layer in the backbone network of
                size: [batch_size, in_channels, conv_h, conv_w])

        Returns a dictionary of Tensors with the following sizes:
            bbox_coords (Tensor): [batch_size, conv_h*conv_w*num_priors, 4]
            class_confs (Tensor): [batch_size, conv_h*conv_w*num_priors, num_classes]
            mask_output (Tensor): [batch_size, conv_h*conv_w*num_priors, mask_dim]
            prior_boxes (Tensor): [conv_h*conv_w*num_priors, 4]
        """
        src = self if self.parent[0] is None else self.parent[0]
        conv_h = inputs.size(2)
        conv_w = inputs.size(3)

        inputs = src.upfeature(inputs)

        bbox_x = src.bbox_extra(inputs)
        conf_x = src.conf_extra(inputs)
        mask_x = src.mask_extra(inputs)

        bbox = (
            src.bbox_layer(bbox_x)
            .permute(0, 2, 3, 1)
            .contiguous()
            .view(inputs.size(0), -1, 4)
        )
        conf = (
            src.conf_layer(conf_x)
            .permute(0, 2, 3, 1)
            .contiguous()
            .view(inputs.size(0), -1, self.num_classes)
        )
        mask = (
            src.mask_layer(mask_x)
            .permute(0, 2, 3, 1)
            .contiguous()
            .view(inputs.size(0), -1, self.mask_dim)
        )

        mask = torch.tanh(mask)

        priors = self.make_priors(conv_h, conv_w)
        preds = {"loc": bbox, "conf": conf, "mask": mask, "priors": priors}
        return preds

    def make_priors(self, conv_h: int, conv_w: int) -> Optional[Tensor]:
        """Priors are [center-x, center-y, width, height] where center-x and
        center-y are the center coordinates of the box.

        Args:
            conv_h (int): The height of the convolutional output.
            conv_w (int): The width of the convolutional output.

        Returns:
            self.priors (Tensor): [conv_h * conv_w * num_priors, 4]
        """
        if self.last_conv_size != (conv_w, conv_h):
            prior_data = []
            for j, i in product(range(conv_h), range(conv_w)):
                # +0.5 because priors are in center-size notation
                center_x = (i + 0.5) / conv_w
                center_y = (j + 0.5) / conv_h
                for scale, aspect_ratios in zip(self.scales, self.aspect_ratios):  # type: ignore
                    for aspect_ratio in aspect_ratios:
                        aspect_ratio = sqrt(aspect_ratio)
                        width = scale * aspect_ratio / self.input_size
                        height = width
                        prior_data += [center_x, center_y, width, height]
            self.priors = Tensor(prior_data).view(-1, 4)
            self.last_conv_size = (conv_w, conv_h)
        return self.priors


class FPNPhase1(ScriptModuleWrapper):
    """First phase of the feature pyramid network"""

    __constants__ = ["interpolation_mode"]

    def __init__(self, in_channels: List[int]) -> None:
        super().__init__()
        self.src_channels = in_channels
        self.lat_layers = nn.ModuleList(
            [
                nn.Conv2d(x, FPN_NUM_FEATURES, kernel_size=1)
                for x in reversed(in_channels)
            ]
        )
        self.interpolation_mode = "bilinear"

    def forward(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        x_1: Optional[torch.Tensor] = None,
        x_2: Optional[torch.Tensor] = None,
        x_3: Optional[torch.Tensor] = None,
        x_4: Optional[torch.Tensor] = None,
        x_5: Optional[torch.Tensor] = None,
        x_6: Optional[torch.Tensor] = None,
        x_7: Optional[torch.Tensor] = None,
    ) -> List[Tensor]:
        """
        Args:
            - convouts (List): A list of convouts for the corresponding layers
                in in_channels.
        Returns:
            - out (List(Tensor)): A list of FPN convouts in the same order as x
                with extra downsample layers if requested.
        """
        convouts_ = [x_1, x_2, x_3, x_4, x_5, x_6, x_7]
        convouts = []

        for count, _ in enumerate(convouts_):
            if convouts_[count] is not None:
                convouts.append(convouts_[count])

        out = []
        lat_feats = []
        x_0 = torch.zeros(1)

        for i in range(len(convouts)):
            out.append(x_0)
            lat_feats.append(x_0)

        count = len(convouts)
        for lat_layer in self.lat_layers:
            count -= 1
            if count < len(convouts) - 1:
                _, _, height, weight = convouts[count].size()  # type: ignore
                x_0 = F.interpolate(
                    x_0,
                    size=(height, weight),
                    mode=self.interpolation_mode,
                    align_corners=False,
                )
            lat_iter = lat_layer(convouts[count])
            lat_feats[count] = lat_iter

            x_0 = x_0 + lat_iter

            out[count] = x_0

        for i in range(len(convouts)):
            out.append(lat_feats[i])
        return out


class FPNPhase2(ScriptModuleWrapper):
    """Second phase of the feature pyramid network"""

    __constants__ = ["num_downsample"]

    def __init__(self, in_channels: List[int]) -> None:
        super().__init__()
        self.src_channels = in_channels
        self.pred_layers = nn.ModuleList(
            [
                nn.Conv2d(FPN_NUM_FEATURES, FPN_NUM_FEATURES, kernel_size=3, padding=1)
                for _ in in_channels
            ]
        )
        self.num_downsample = NUM_DOWNSAMPLE
        self.downsample_layers = nn.ModuleList(
            [
                nn.Conv2d(
                    FPN_NUM_FEATURES,
                    FPN_NUM_FEATURES,
                    kernel_size=3,
                    padding=1,
                    stride=2,
                )
                for _ in range(self.num_downsample)
            ]
        )

    def forward(  # pylint: disable=too-many-arguments
        self,
        x_1: Optional[List] = None,
        x_2: Optional[List] = None,
        x_3: Optional[List] = None,
        x_4: Optional[List] = None,
        x_5: Optional[List] = None,
        x_6: Optional[List] = None,
        x_7: Optional[List] = None,
    ) -> List[Optional[List[Any]]]:
        """
        Args:
            x_1, x_2, x_3, x_4, x_5, x_6 (List): A list of convouts for the
                corresponding layers in in_channels.

        Returns:
            out (List): A list of FPN convouts in the same order as x with extra
                downsample layers if requested.
        """
        out_ = [x_1, x_2, x_3, x_4, x_5, x_6, x_7]
        out = []
        for count, _ in enumerate(out_):
            if out_[count] is not None:
                out.append(out_[count])

        j = len(out)
        for pred_layer in self.pred_layers:
            j -= 1
            out[j] = F.relu(pred_layer(out[j]))  # type: ignore
        for downsample_layer in self.downsample_layers:
            out.append(downsample_layer(out[-1]))
        return out


class YolactEdgeHead:
    """
    This is the final layer of Single Shot Detection (SSD). Decode location preds,
    apply non-maximum suppression to location predictions based on conf scores and
    threshold to a top maximum number output predictions for both confidence scores
    and locations, as the predicted masks.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        num_classes: int,
        bkg_label: int,
        conf_thresh: float,
        iou_threshold: float,
        max_num_detections: int,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_classes = num_classes
        self.background_label = bkg_label
        self.iou_threshold = iou_threshold
        self.conf_thresh = conf_thresh
        self.max_num_detections = max_num_detections

    def __call__(self, predictions: Dict[str, torch.Tensor]) -> List[Any]:
        """
        Args:
            loc_data: (tensor) Loc preds from loc layers
                Shape: [batch, num_priors, 4]
            conf_data: (tensor) Shape: Conf preds from conf layers
                Shape: [batch, num_priors, num_classes]
            mask_data: (tensor) Mask preds from mask layers
                Shape: [batch, num_priors, mask_dim]
            prior_data: (tensor) Prior boxes and variances from priorbox layers
                Shape: [num_priors, 4]
            proto_data: (tensor) If using mask_type.lincomb, the prototype masks
                Shape: [batch, mask_h, mask_w, mask_dim]

        Returns:
            out (List): output of shape (batch_size, max_num_detections, 1 + 1 + 4 + mask_dim)
                These outputs are in the order: class idx, confidence, bbox coords,
                and mask.
        """
        loc_data = predictions["loc"]
        conf_data = predictions["conf"]
        mask_data = predictions["mask"]
        prior_data = predictions["priors"]
        proto_data = predictions["proto"] if "proto" in predictions else None

        out = []
        batch_size = loc_data.size(0)
        num_priors = prior_data.size(0)
        conf_preds = (
            conf_data.view(batch_size, num_priors, self.num_classes)
            .transpose(2, 1)
            .contiguous()
        )

        for batch_idx in range(batch_size):
            decoded_boxes = decode(loc_data[batch_idx], prior_data)
            result = self.detect(batch_idx, conf_preds, decoded_boxes, mask_data)
            if result is not None and proto_data is not None:
                result["proto"] = proto_data[batch_idx]
            out.append(result)
        return out

    def detect(  # pylint: disable=too-many-arguments
        self,
        batch_idx: int,
        conf_preds: Tensor,
        decoded_boxes: Tensor,
        mask_data: Tensor,
    ) -> Optional[Dict[str, Any]]:
        """Perform nms for only the max scoring class that isn't background (class 0)

        Args:
            batch_idx: (int) The batch index
            conf_preds: (Tensor) Confidence predictions for each prior
            decoded_boxes: (Tensor) Decoded boxes for each prior
            mask_data: (Tensor) Mask predictions for each prior

        Returns:
            A dictionary containing the following keys:
                box (Tensor): The bounding box values for each detection (0 to 1)
                mask (Tensor): The segmentation mask for each detection (-1 to 1)
                class (Tensor): Class ID for each detection (0 to 80)
                score (Tensor): Confidence score for each detection (0 to 1)
        """
        cur_scores = conf_preds[batch_idx, 1:, :]
        conf_scores, _ = torch.max(cur_scores, dim=0)
        keep = conf_scores > self.conf_thresh
        scores = cur_scores[:, keep]
        boxes = decoded_boxes[keep, :]
        masks = mask_data[batch_idx, keep, :]

        if scores.size(1) == 0:
            return None

        boxes, masks, classes, scores = self.fast_nms(
            boxes,
            masks,
            scores,
            self.iou_threshold,
            self.max_num_detections,
        )

        return {"box": boxes, "mask": masks, "class": classes, "score": scores}

    @classmethod
    def fast_nms(  # pylint: disable=too-many-arguments
        cls,
        boxes: Tensor,
        masks: Tensor,
        scores: Tensor,
        iou_threshold: float,
        max_num_detections: int,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """Non-maximum Suppression

        Args:
            boxes (Tensor): Bounding boxes for each object
            masks (Tensor): Masks for each object
            scores (Tensor): Confidence scores for each object
            iou_threshold (float): IoU threshold for NMS
            max_num_detections (int): Maximum number of detections

        Returns:
            boxes (np.ndarray): array of detected bboxes
            masks (np.ndarray): array of detected masks
            classes (np.ndarray): array of class labels
            scores (np.ndarray): array of detection confidence scores
        """
        scores, idx = scores.sort(1, descending=True)
        idx = idx[:, :max_num_detections].contiguous()
        scores = scores[:, :max_num_detections]
        num_classes, num_dets = idx.size()
        boxes = boxes[idx.view(-1), :].view(num_classes, num_dets, 4)
        masks = masks[idx.view(-1), :].view(num_classes, num_dets, -1)

        iou = jaccard(boxes, boxes)
        iou.triu_(diagonal=1)
        iou_max, _ = iou.max(dim=1)

        # Filter out the ones that are higher that the IoU threshold
        keep = iou_max <= iou_threshold

        classes = torch.arange(num_classes, device=boxes.device)[:, None].expand_as(
            keep
        )
        classes = classes[keep]

        boxes = boxes[keep]
        masks = masks[keep]
        scores = scores[keep]

        scores, idx = scores.sort(0, descending=True)
        idx = idx[:max_num_detections]
        scores = scores[:max_num_detections]
        classes = classes[idx]
        boxes = boxes[idx]
        masks = masks[idx]

        return boxes, masks, classes, scores



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions to map human-friendly class labels to detect IDs."""

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import yaml

# Master map file for class name to object IDs for object detection models
MASTER_MAP = "pipeline/nodes/model/master_map.yml"

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def obj_det_load_class_id_mapping(node_name: str) -> Dict[str, int]:
    """Loads class name to object ID mapping from the file
    peekingduck/pipeline/nodes/model/master_map.yml
    for object detection models only.

    Tech Notes
        master_map.yml comprises two documents:
        1. model mapping: tells which mapping system a particular object detection model uses
        2. class name to ID mapping: maps class name to ID, supports multiple mapping systems

    Args:
        node_name (str): Tells function which mapping to load,
                            Possible values = { model.efficientdet, model.yolo, model.yolox }.

    Returns:
        Dict[str, int]: Mapping of class names to object IDs relevant to given node_name
    """
    # use __file__ instead of self._base_dir as latter can be set to any (temp) path
    # without `peekingduck/` subdirectory, resulting in master map file not found error
    master_map_file = Path(__file__).parents[1] / MASTER_MAP
    # read both documents from master_map.yml
    with master_map_file.open() as map_file:
        model_map_dict, class_id_map_dict = yaml.safe_load_all(map_file)

    # node_name sanity check, to preempt non-object detection model nodes
    node_name_list = list(map(lambda x: f"model.{x}", model_map_dict.keys()))
    assert (
        node_name in node_name_list
    ), f"Node Name Error: expect one of {node_name_list} but got {node_name}"

    # determine mapping system to use based on given model
    model = node_name.replace("model.", "")
    model_map = model_map_dict[model]

    # construct class name->object ID map for the model
    class_id_map = {}
    for key, val in class_id_map_dict.items():
        class_id = val[model_map]
        class_id_map[key.lower()] = class_id
    return class_id_map


def obj_det_change_class_name_to_id(
    node_name: str, key: str, value: List[Any]
) -> Tuple[str, List[int]]:
    """Process object detection model node's detect key and check for
    any class names to be converted to object IDs.
    E.g. person to 0, car to 2

    Args:
        node_name (str): to determine which object detection model is being used
                         because different models can use different object IDs.
        key (str): expected to be "detect"; error otherwise.
        value (List[Any]): list of class names or object IDs for detection.
                           If object IDs, do nothing.
                           If class names, convert to object IDs.

    Returns:
        Tuple[str, List[int]]: "detect", list of sorted object IDs.
    """
    class_id_map = obj_det_load_class_id_mapping(node_name)

    if not value:
        logger.warning("detect list is empty, defaulting to detect person")
        value = ["person"]
    elif value == ["*"]:
        logger.info("Detecting all object classes")
        value = [*class_id_map]

    value_lc = [x.lower() if isinstance(x, str) else x for x in value]

    # parse value_lc for possible class name errors
    invalid_class_names = []
    for class_name in value_lc:
        if isinstance(class_name, str) and class_name not in class_id_map:
            invalid_class_names.append(class_name)

    if invalid_class_names:
        logger.warning(f"Invalid class names: {invalid_class_names}")

    # convert class names to numeric object IDs, any errors default to zero
    obj_ids_set = {
        x if isinstance(x, int) else class_id_map.get(x, 0) for x in value_lc
    }
    obj_ids_sorted_list = sorted(list(obj_ids_set))
    return key, obj_ids_sorted_list



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Calculates camera coefficients to be used to
remove distortion from a wide-angle camera image.
"""

import math
import time
from pathlib import Path
from typing import Any, Dict, List

import cv2
import numpy as np
import yaml

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.constants import BLACK, CHAMPAGNE, TOMATO

# global constants
# terminal criteria for subpixel finetuning
TERMINATION_CRITERIA = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

NUM_PICTURES = 5

# constants for positioning of boxes and text
TOP_LEFT = 0
TOP_RIGHT = 1
BOTTOM_LEFT = 2
BOTTOM_RIGHT = 3
MIDDLE = 4

# constants for _check_corners_validity
AREA_THRESHOLD = 3 / 4
CORNERS_OK = 0
IMAGE_TOO_SMALL = 1
NOT_IN_BOX = 2

# displayed messages
DEFAULT_TEXT = ["PLACE BOARD HERE"]
TOO_SMALL = ["MOVE BOARD CLOSER"]
DETECTION_SUCCESS = ["DETECTION SUCCESSFUL!", "PRESS ANY KEY TO CONTINUE."]
DETECTION_COMPLETE = ["DETECTION COMPLETE!", "PRESS ANY KEY TO EXIT."]
MAX_LEN_TEXT = "PRESS ANY KEY TO CONTINUE."

# constants for drawing
BGND_BOX_OPACITY = 0.75
BOX_WIDTH_RATIO = 1 / 3  # relative to the window width
BOX_HEIGHT_RATIO = 1 / 2  # relative to the window height
TEXT_PADDING = 5

# constants for font drawing
# 1280 is used as the reference point for the ratio
BOX_THICKNESS_RATIO = 2 / 1280
NORMAL_FONT_SCALE_RATIO = 0.84 / 1280
COUNTDOWN_FONT_SCALE_RATIO = 10 / 1280
NORMAL_FONT_THICKNESS_RATIO = 2 / 1280
COUNTDOWN_FONT_THICKNESS_RATIO = 8 / 1280


class Node(AbstractNode):
    """Calculates camera coefficients for `undistortion
    <https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html>`_.

    To calculate your camera, first download the following checkerboard and print it
    out in a suitable size and attach it to a hard surface, or display it on a sufficiently
    large device screen, such as a computer or a tablet. For most use cases, an A4-sized
    checkerboard works well, but depending on the position and distance of the camera, a
    bigger checkerboard may be required.

    .. image:: /assets/api/checkerboard.png
        :width: 20 %

    Next, create an empty ``pipeline_config.yml`` in your project folder and modify it as follows:

    .. code-block:: yaml
        :linenos:

        nodes:
        - input.visual:
            source: 0 # change this to the camera you are using
            threading: True
            mirror_image: True
        - dabble.camera_calibration
        - output.screen

    Run the above pipeline with :greenbox:`peekingduck run`. If you are unfamiliar with the pipeline
    file and running peekingduck, you may refer to the
    :doc:`HelloCV tutorial </tutorials/01_hello_cv>`. |br|
    You should see a display of your camera with some instructions overlaid. Follow the instructions
    to position the checkerboard at 5 different positions in the camera. If the process is
    successful, the camera coefficients will be calculated and written to a file and you can start
    using the :mod:`augment.undistort` node.

    Inputs:
        |img_data|

    Outputs:
        |img_data|


    Configs:
        num_corners (:obj:`List[int]`):
            **default = [10, 7]**. |br|
            A list containing the number of internal corners along the vertical
            and horizontal axes. For example, in the given image above, the
            checkerboard is of size 11x8, so the number of internal corners is
            10x7. If you are using the given checkerboard above, you do not need
            to change this parameter.
        scale_factor (:obj:`int`):
            **default = 2**. |br|
            Factor to scale the image by when finding chessboard corners. For
            example, with a scale of 4, an image of size (1080 x 1920) will be
            scaled down to (270 x 480) when detecting the corners. Increasing this
            value reduces computation time. If the node is unable to detect corners,
            reducing this value may help.
        file_path (:obj:`str`):
            **default = "PeekingDuck/data/camera_calibration_coeffs.yml"**. |br|
            Path of the YML file to store the calculated camera coefficients.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.file_path = Path(self.file_path)  # type: ignore
        # check if file_path has a ".yml" extension
        if self.file_path.suffix != ".yml":
            raise ValueError("Filepath must have a '.yml' extension.")
        if not self.file_path.exists():
            self.file_path.parent.mkdir(parents=True, exist_ok=True)

        grid_height = self.num_corners[0]
        grid_width = self.num_corners[1]

        # prepare all object points, like (0, 0, 0), (1, 0, 0), etc.
        np_mgrid = np.mgrid[0:grid_height, 0:grid_width]
        object_points_base = np.zeros((grid_height * grid_width, 3), np.float32)
        object_points_base[:, :2] = np_mgrid.T.reshape(-1, 2)

        # arrays to store object points and image points
        # points in real world
        self.object_points: List[np.ndarray] = [object_points_base] * NUM_PICTURES
        self.image_points: List[np.ndarray] = []  # points on image plane

        self.last_detection = time.time()
        self.num_detections = 0

        self.display_scales: Dict["str", Any["float", "int"]] = {}

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:  # type: ignore
        """This node calculates the camera distortion coefficients for undistortion.

        Args:
            inputs (dict): Inputs dictionary with the key `img`.

        Returns:
            outputs (dict): Outputs dictionary with the key `img`.
        """

        img = inputs["img"]
        gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        height, width = img.shape[:2]

        self._initialize_display_scales(width)
        start_point, end_point, text_pos = _get_box_info(
            self.num_detections, width, height
        )

        detect_corners_success = False

        _draw_box(img, start_point, end_point, self.display_scales["box_thickness"])
        text_to_draw = [f"{DEFAULT_TEXT[0]} ({self.num_detections+1}/{NUM_PICTURES})"]

        # if sufficient time has passed, attempt to detect corners
        if time.time() - self.last_detection >= 5:
            detect_corners_success, corners = self._detect_corners(
                height, width, gray_img
            )

        # cv2 successfully detected the corners
        if detect_corners_success:
            corners_valid = _check_corners_validity(
                width, height, corners, start_point, end_point
            )

            if corners_valid == IMAGE_TOO_SMALL:
                text_to_draw = TOO_SMALL

        if detect_corners_success and corners_valid == CORNERS_OK:

            self.image_points.append(corners)
            self.num_detections += 1

            self._draw_text_and_corners(img, gray_img, corners, text_pos)

            # wait for user to press a key
            cv2.waitKey(0)
            self.last_detection = time.time()

            # if we have sufficient images, calculate the coefficients and write to a file
            if self.num_detections == NUM_PICTURES:
                calibration_data = self._calculate_coeffs(
                    img_shape=gray_img.shape[::-1]
                )
                self._write_coeffs(calibration_data)
                self._calculate_error(calibration_data)
                return {"pipeline_end": True}

        self._draw_text_and_countdown(img, text_to_draw, text_pos)

        return {"img": img}

    def _get_config_types(self) -> Dict[str, Any]:
        """Returns dictionary mapping the node's config keys to respective types."""
        return {"num_corners": List[int], "scale_factor": int, "file_path": str}

    def _initialize_display_scales(self, img_width: int) -> None:
        """Initializes display scales if it hasn't been initialized before"""
        if not self.display_scales:
            self.display_scales = {
                "box_thickness": max(int(img_width * BOX_THICKNESS_RATIO), 1),
                "normal_font_scale": img_width * NORMAL_FONT_SCALE_RATIO,
                "countdown_font_scale": img_width * COUNTDOWN_FONT_SCALE_RATIO,
                "normal_font_thickness": int(img_width * NORMAL_FONT_THICKNESS_RATIO),
                "countdown_font_thickness": int(
                    img_width * COUNTDOWN_FONT_THICKNESS_RATIO
                ),
            }

    def _detect_corners(self, height: int, width: int, gray_img: np.ndarray) -> tuple:
        """Detects corners in the image"""
        # downscale
        new_h = int(height / self.scale_factor)
        new_w = int(width / self.scale_factor)

        resized_img = cv2.resize(gray_img, (new_w, new_h), interpolation=cv2.INTER_AREA)

        # try to find chessboard corners
        detect_corners_success, corners = cv2.findChessboardCorners(
            image=resized_img, patternSize=self.num_corners, corners=None
        )

        if corners is not None:
            corners = corners * self.scale_factor

        return detect_corners_success, corners

    def _calculate_coeffs(self, img_shape: tuple) -> tuple:
        """Performs calculations with detected corners"""
        calibration_data = cv2.calibrateCamera(
            objectPoints=self.object_points,
            imagePoints=self.image_points,
            imageSize=img_shape,
            cameraMatrix=None,
            distCoeffs=None,
        )

        (
            calibration_success,
            camera_matrix,
            distortion_coeffs,
            _,
            _,
        ) = calibration_data

        if calibration_success:
            self.logger.info("Calibration successful!")
            self.logger.info(f"Camera Matrix: {camera_matrix}")
            self.logger.info(f"Distortion Coefficients: {distortion_coeffs}")
        else:
            raise Exception("Calibration failed. Please try again.")

        return calibration_data

    def _write_coeffs(self, calibration_data: tuple) -> None:
        """Writes camera coefficients to a file"""
        (_, camera_matrix, distortion_coeffs, _, _) = calibration_data

        file_data = {}
        file_data["camera_matrix"] = camera_matrix.tolist()
        file_data["distortion_coeffs"] = distortion_coeffs.tolist()

        yaml.dump(file_data, open(self.file_path, "w"), default_flow_style=None)

    def _calculate_error(self, calibration_data: tuple) -> None:
        """Calculates re-projection error"""
        (
            _,
            camera_matrix,
            distortion_coeffs,
            rotation_vec,
            translation_vec,
        ) = calibration_data

        mean_error = 0
        for i in range(len(self.object_points)):
            projected_image_points, _ = cv2.projectPoints(
                objectPoints=self.object_points[i],
                rvec=rotation_vec[i],
                tvec=translation_vec[i],
                cameraMatrix=camera_matrix,
                distCoeffs=distortion_coeffs,
            )

            error = cv2.norm(
                src1=self.image_points[i],
                src2=projected_image_points,
                normType=cv2.NORM_L2,
            )
            error /= len(projected_image_points)
            mean_error += error

        self.logger.info(f"Total error: {mean_error / len(self.object_points)}")

    def _draw_text_and_corners(
        self,
        img: np.ndarray,
        gray_img: np.ndarray,
        corners: np.ndarray,
        text_pos: tuple,
    ) -> None:
        """Draws text and corners on image"""
        # improve corner accuracy
        corners_accurate = cv2.cornerSubPix(
            image=gray_img,
            corners=corners,
            winSize=(11, 11),
            zeroZone=(-1, -1),
            criteria=TERMINATION_CRITERIA,
        )

        # draw corners and message on the image
        cv2.drawChessboardCorners(
            image=img,
            patternSize=self.num_corners,
            corners=corners_accurate,
            patternWasFound=True,
        )

        if self.num_detections != NUM_PICTURES:
            text_to_draw = DETECTION_SUCCESS
        else:
            text_to_draw = DETECTION_COMPLETE

        _draw_text(
            img=img,
            texts=text_to_draw,
            pos_info=text_pos,
            font_scale=self.display_scales["normal_font_scale"],
            thickness=self.display_scales["normal_font_thickness"],
        )

        # display the image
        cv2.imshow("PeekingDuck", img)

    def _draw_text_and_countdown(
        self, img: np.ndarray, text_to_draw: List[str], text_pos: tuple
    ) -> None:
        """Draws text and countdown on image"""
        _draw_text(
            img=img,
            texts=text_to_draw,
            pos_info=text_pos,
            font_scale=self.display_scales["normal_font_scale"],
            thickness=self.display_scales["normal_font_thickness"],
        )

        time_to_next_detection = math.ceil(5 - time.time() + self.last_detection)
        if time_to_next_detection > 0:
            _draw_countdown(
                img=img,
                num=time_to_next_detection,
                font_scale=self.display_scales["countdown_font_scale"],
                thickness=self.display_scales["countdown_font_thickness"],
            )


def _get_box_info(num: int, width: int, height: int) -> tuple:
    """Returns start and end points of box, and position to put text"""
    start_points = {
        TOP_LEFT: (0, 0),
        TOP_RIGHT: (int(width * (1 - BOX_WIDTH_RATIO)), 0),
        BOTTOM_LEFT: (0, int(height * (1 - BOX_HEIGHT_RATIO))),
        BOTTOM_RIGHT: (
            int(width * (1 - BOX_WIDTH_RATIO)),
            int(height * (1 - BOX_HEIGHT_RATIO)),
        ),
        MIDDLE: (
            int(width * (1 / 2 - BOX_WIDTH_RATIO / 2)),
            int(height * (1 / 2 - BOX_HEIGHT_RATIO / 2)),
        ),
    }
    end_points = {
        TOP_LEFT: (int(width * BOX_WIDTH_RATIO), int(height * BOX_HEIGHT_RATIO)),
        TOP_RIGHT: (width, int(height * BOX_HEIGHT_RATIO)),
        BOTTOM_LEFT: (int(width * BOX_WIDTH_RATIO), height),
        BOTTOM_RIGHT: (width, height),
        MIDDLE: (
            int(width * (1 / 2 + BOX_WIDTH_RATIO / 2)),
            int(height * (1 / 2 + BOX_HEIGHT_RATIO / 2)),
        ),
    }
    text_positions = {
        TOP_LEFT: (TEXT_PADDING, int(height * BOX_HEIGHT_RATIO) - TEXT_PADDING),
        TOP_RIGHT: (
            int(width * (1 - BOX_WIDTH_RATIO)) + TEXT_PADDING,
            int(height * BOX_HEIGHT_RATIO) - TEXT_PADDING,
        ),
        BOTTOM_LEFT: (
            TEXT_PADDING,
            int(height * (1 - BOX_HEIGHT_RATIO)) + TEXT_PADDING,
        ),
        BOTTOM_RIGHT: (
            int(width * (1 - BOX_WIDTH_RATIO)) + TEXT_PADDING,
            int(height * (1 - BOX_HEIGHT_RATIO)) + TEXT_PADDING,
        ),
        MIDDLE: (
            int(width * (1 / 2 - BOX_WIDTH_RATIO / 2)) + TEXT_PADDING,
            int(height * (1 / 2 + BOX_HEIGHT_RATIO / 2)) - TEXT_PADDING,
        ),
    }
    pos_types = {
        TOP_LEFT: BOTTOM_LEFT,
        TOP_RIGHT: BOTTOM_LEFT,
        BOTTOM_LEFT: TOP_LEFT,
        BOTTOM_RIGHT: TOP_LEFT,
        MIDDLE: BOTTOM_LEFT,
    }

    return start_points[num], end_points[num], (text_positions[num], pos_types[num])


def _check_corners_validity(
    width: int, height: int, corners: np.ndarray, start_point: tuple, end_point: tuple
) -> int:
    """Checks whether the corners are large enough and fall within the box"""
    min_w = width
    min_h = height
    max_w = 0
    max_h = 0
    for corner in corners:
        min_w = min(min_w, corner[0][0])
        max_w = max(max_w, corner[0][0])
        min_h = min(min_h, corner[0][1])
        max_h = max(max_h, corner[0][1])

    area = (max_w - min_w) * (max_h - min_h)

    # if area is less than 1/4 of the box size
    if area < width * BOX_WIDTH_RATIO * height * BOX_HEIGHT_RATIO / 4:
        return IMAGE_TOO_SMALL

    # if the board is completely out of the box
    if (
        max_w < start_point[0]
        or end_point[0] < min_w
        or max_h < start_point[1]
        or end_point[1] < min_h
    ):
        return NOT_IN_BOX

    min_w_box = max(min_w, start_point[0])
    max_w_box = min(max_w, end_point[0])
    min_h_box = max(min_h, start_point[1])
    max_h_box = min(max_h, end_point[1])

    # check if at least 3 / 4 of the board area is within the box
    if (max_w_box - min_w_box) * (max_h_box - min_h_box) < area * AREA_THRESHOLD:
        return NOT_IN_BOX

    return CORNERS_OK


def _draw_box(
    img: np.ndarray, start_point: tuple, end_point: tuple, box_thickness: int
) -> None:
    """Draws rectangle on the image"""
    cv2.rectangle(
        img=img,
        pt1=start_point,
        pt2=end_point,
        color=(0, 0, 0),
        thickness=3 * box_thickness,
    )

    cv2.rectangle(
        img=img,
        pt1=start_point,
        pt2=end_point,
        color=CHAMPAGNE,
        thickness=box_thickness,
    )


def _draw_bgnd_box(
    img: np.ndarray,
    pt1: tuple,
    pt2: tuple,
) -> None:
    """Draws background box on image"""
    box_img = img.copy()

    # draw the rectangle
    cv2.rectangle(box_img, pt1, pt2, BLACK, cv2.FILLED)

    # apply the overlay
    cv2.addWeighted(box_img, BGND_BOX_OPACITY, img, 1 - BGND_BOX_OPACITY, 0, img)


def _draw_text(
    img: np.ndarray,
    texts: List[str],
    pos_info: tuple,
    font_scale: float,
    thickness: int,
) -> None:
    """Draws text on the image"""
    pos, pos_type = pos_info

    text_width = 0
    (_, text_height), baseline = cv2.getTextSize(
        text=texts[0],
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=font_scale,
        thickness=thickness,
    )

    for text in texts:
        text_width = max(
            text_width,
            cv2.getTextSize(
                text=text,
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=font_scale,
                thickness=thickness,
            )[0][0],
        )

    img_width = img.shape[1]
    box_width = min(text_width + 10, int(img_width * BOX_WIDTH_RATIO - 10))
    if pos_type == TOP_LEFT:
        _draw_bgnd_box(
            img,
            (pos[0], pos[1]),
            (
                pos[0] + box_width,
                pos[1] + len(texts) * (text_height + baseline + 5) + baseline + 5,
            ),
        )

        pos = (pos[0] + 5, pos[1] + text_height + baseline + 5)

    elif pos_type == BOTTOM_LEFT:
        _draw_bgnd_box(
            img,
            (pos[0], pos[1] - len(texts) * (text_height + baseline + 5) - baseline - 5),
            (pos[0] + box_width, pos[1]),
        )

        pos = (
            pos[0] + 5,
            pos[1] - (text_height + baseline + 5) * (len(texts) - 1) - baseline - 5,
        )

    for text in texts:
        cv2.putText(
            img=img,
            text=text,
            org=pos,
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=font_scale,
            color=CHAMPAGNE,
            thickness=thickness,
            lineType=cv2.LINE_AA,
        )

        pos = (pos[0], pos[1] + text_height + baseline + 5)


def _draw_countdown(
    img: np.ndarray, num: int, font_scale: float, thickness: int
) -> None:
    """Draws a countdown in the center of the screen"""
    height, width = img.shape[:2]

    text = str(num)

    text_width, text_height = cv2.getTextSize(
        text=text,
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=font_scale,
        thickness=thickness,
    )[0]

    pos = (int(width / 2 - text_width / 2), int(height / 2 + text_height / 2))

    cv2.putText(
        img=img,
        text=text,
        org=pos,
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=font_scale,
        color=BLACK,
        thickness=thickness * 3,
        lineType=cv2.LINE_AA,
    )

    cv2.putText(
        img=img,
        text=text,
        org=pos,
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=font_scale,
        color=TOMATO,
        thickness=thickness,
        lineType=cv2.LINE_AA,
    )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the PeekingDuck Pipeline PlayList class"""

from typing import Dict, List, Union
from datetime import datetime
from pathlib import Path
import logging
import yaml


# Globals
PKD_CONFIG_DIR = ".peekingduck"
PKD_PLAYLIST_FILE = "playlist.yml"


class PipelineStats:
    """Implements immutable PipelineStats class to store pipeline-related information."""

    def __init__(self, pipeline: str) -> None:
        self._hash = hash(pipeline)
        self._pipeline = pipeline
        pipeline_path = Path(pipeline)
        self._name = pipeline_path.name
        self._datetime = (
            pipeline_path.stat().st_mtime if pipeline_path.exists() else None
        )

    def __eq__(self, obj: "PipelineStats") -> bool:
        return self._pipeline == obj._pipeline

    def __hash__(self) -> int:
        return self._hash

    def __lt__(self, obj: "PipelineStats") -> bool:
        return (self._name, self._pipeline) < (obj._name, obj._pipeline)

    def __repr__(self) -> str:
        return self._pipeline

    @property
    def datetime(self) -> str:
        """Get last modified date/time of pipeline file.

        Returns:
            str: Last modified date/time string.
        """
        if self._datetime:
            return datetime.fromtimestamp(self._datetime).strftime("%Y-%m-%d-%H:%M:%S")
        return ""

    @property
    def name(self) -> str:
        """Get pipeline name.

        Returns:
            str: Name of pipeline.
        """
        return self._name

    @property
    def pipeline(self) -> str:
        """Get pipeline full path name.

        Returns:
            str: Full path name of pipeline.
        """
        return self._pipeline


class PlayList:
    """Implements the PlayList class to store pipelines in playlist and to handle
    the internal data structures for managing pipelines"""

    def __init__(self, home_path: Path) -> None:
        self.logger = logging.getLogger(__name__)
        # Construct path to ~user_home/.peekingduck/playlist.yaml
        self._playlist_dir = home_path / PKD_CONFIG_DIR
        self._playlist_dir.mkdir(exist_ok=True)
        self._playlist_path = self._playlist_dir / PKD_PLAYLIST_FILE
        self.logger.debug(f"playlist_path={self._playlist_path}")
        self.load_playlist_file()

    def __iter__(self) -> "PlayList":
        self._iter_idx = -1
        return self

    def __next__(self) -> PipelineStats:
        self._iter_idx += 1
        if self._iter_idx < len(self.pipeline_stats):
            return self.pipeline_stats[self._iter_idx]
        raise StopIteration

    def __contains__(self, item: str) -> bool:
        if not isinstance(item, str):
            item = str(item)
        res = False
        for pipeline_stats in self.pipeline_stats:
            if pipeline_stats.pipeline == item:
                res = True
                break
        return res

    def __getitem__(self, key: str) -> PipelineStats:
        return self._pipelines_dict[key]

    def __len__(self) -> int:
        return len(self.pipeline_stats)

    #
    # Internal methods
    #
    def _read_playlist_file(self) -> List[str]:
        """Read contents of playlist file, if any

        Returns:
            List[str]: contents of playlist file, a list of pipelines
        """
        if not Path.exists(self._playlist_path):
            self.logger.debug(f"{self._playlist_path} not found")
            return []

        with open(self._playlist_path, "r", encoding="utf-8") as file:
            playlist = yaml.safe_load(file)

        return playlist["playlist"]

    #
    # External methods
    #
    def add_pipeline(self, pipeline_path: Union[Path, str]) -> None:
        """Add pipeline yaml file to playlist.
           Do nothing if pipeline is already in playlist.

        Args:
            pipeline_path (Union[Path, str]): path of yaml file to add
        """
        pipeline_str = str(pipeline_path)
        if pipeline_str in self:
            self.logger.info(f"{pipeline_str} already in playlist")
            return
        pipeline_stats = PipelineStats(pipeline_str)
        self.pipeline_stats.append(pipeline_stats)
        self._pipelines_dict[pipeline_str] = pipeline_stats

    def delete_pipeline(self, pipeline_path: Union[Path, str]) -> None:
        """Delete pipeline yaml file from playlist.
           Do nothing if pipeline is not in playlist.

        Args:
            pipeline_path (Union[Path, str]): path of yaml file to delete
        """
        pipeline_str = str(pipeline_path)
        if pipeline_str in self:
            pipeline_stats = PipelineStats(pipeline_str)
            self.pipeline_stats.remove(pipeline_stats)
            self._pipelines_dict.pop(pipeline_str)

    def load_playlist_file(self) -> None:
        """Load playlist file"""
        pipelines = self._read_playlist_file()
        self.pipeline_stats: List[PipelineStats] = []
        self._pipelines_dict: Dict[str, PipelineStats] = {}
        for pipeline in pipelines:
            self.add_pipeline(pipeline)

    def save_playlist_file(self) -> None:
        """Save playlist file"""
        # construct playlist contents with full pathnames
        playlist = [str(stats) for stats in self.pipeline_stats]
        playlist_dict = {"playlist": playlist}
        self.logger.debug(f"playlist_dict={playlist_dict}")

        with open(self._playlist_path, "w", encoding="utf8") as file:
            yaml.dump(playlist_dict, file)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of bboxes and masks for mask_rcnn
"""

import logging
from pathlib import Path
from typing import Dict, List, Tuple
import cv2
import numpy as np
import torch
from torch import Tensor
import torchvision.transforms as T
from peekingduck.pipeline.utils.bbox.transforms import xyxy2xyxyn
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection.backbone_utils import (
    resnet_fpn_backbone,
)
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection.mask_rcnn import (
    MaskRCNN,
)


class Detector:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Detector class to handle detection of bboxes and masks for mask_rcnn"""

    preprocess_transform = T.Compose(
        [
            T.ToTensor(),
            T.ConvertImageDtype(torch.float32),
        ]
    )
    model_name_map = {"r50-fpn": "resnet50", "r101-fpn": "resnet101"}

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: Dict[int, str],
        detect_ids: List[int],
        model_type: str,
        num_classes: int,
        model_file: Dict[str, str],
        min_size: int,
        max_size: int,
        iou_threshold: float,
        max_num_detections: int,
        score_threshold: float,
        mask_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.class_names = class_names
        self.detect_ids = torch.tensor(
            detect_ids, dtype=torch.int64, device=self.device
        )
        self.model_type = model_type
        self.num_classes = num_classes
        self.model_path = model_dir / model_file[self.model_type]
        self.min_size = min_size
        self.max_size = max_size
        self.iou_threshold = iou_threshold
        self.max_num_detections = max_num_detections
        self.score_threshold = score_threshold
        self.mask_threshold = mask_threshold
        self.mask_rcnn = self._create_mask_rcnn_model()
        self.filtered_output: Dict[str, Tensor] = {}

    @torch.no_grad()
    def predict_instance_mask_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Mask R-CNN masks and bboxes prediction function

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): array of detected bboxes
            labels (np.ndarray): array of labels
            scores (np.ndarray): array of scores
            masks (np.ndarray): array of detected masks
        """
        img_shape = image.shape[:2]
        processed_images = self._preprocess(image)
        # run network, assumes batch size of 1 for peekingduck inference
        network_output = self.mask_rcnn(processed_images)[0]
        bboxes, labels, scores, masks = self._postprocess(network_output, img_shape)

        return bboxes, labels, scores, masks

    def _create_mask_rcnn_model(self) -> MaskRCNN:
        """Creates a Mask-RCNN model and loads its weights. It also logs model configurations.

        Returns:
            (MaskRCNN): An initialized and loaded model in PyTorch framework
        """
        self.logger.info(
            "Mask-RCNN model loaded with following configs:\n\t"
            f"Model type: {self.model_type}\n\t"
            f"IDs being detected: {self.detect_ids.tolist()}\n\t"
            f"IOU threshold: {self.iou_threshold}\n\t"
            f"Score threshold: {self.score_threshold}\n\t"
            f"Mask threshold: {self.mask_threshold}\n\t"
            f"Maximum number of detections per image: {self.max_num_detections}\n\t"
            f"Maximum size of the image: {self.max_size}\n\t"
            f"Minimum size of the image: {self.min_size}"
        )

        return self._load_mask_rcnn_weights()

    def _get_model(self) -> MaskRCNN:
        """Constructs Mask-RCNN model based on parsed configuration.

        Returns:
            (MaskRCNN): Mask-RCNN model.
        """
        backbone_name = Detector.model_name_map[self.model_type]
        backbone = resnet_fpn_backbone(
            backbone_name=backbone_name,
        )
        return MaskRCNN(
            backbone=backbone,
            num_classes=self.num_classes,
            box_nms_thresh=self.iou_threshold,
            box_score_thresh=self.score_threshold,
            box_detections_per_img=self.max_num_detections,
            min_size=self.min_size,
            max_size=self.max_size,
        )

    def _load_mask_rcnn_weights(self) -> MaskRCNN:
        """Loads Mask-RCNN model weights

        Raises:
            FileNotFoundError: Raise error when model path does not exist

        Returns:
            (MaskRCNN): Mask-RCNN model loaded with weights
        """
        if self.model_path.is_file():
            state_dict = torch.load(self.model_path, map_location=self.device)
            model = self._get_model()
            model.load_state_dict(state_dict)
            model.eval().to(self.device)
            return model

        raise FileNotFoundError(
            f"Model file does not exist. Please check that {self.model_path} exists."
        )

    def _postprocess(
        self,
        network_output: Dict[str, Tensor],
        img_shape: Tuple[int, int],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Postprocessing of detected bboxes and masks for mask_rcnn

        Args:
            network_output (Dict[str, Tensor]): A dictionary from the first element of the
                Mask-RCNN output. The keys are should have "labels", "boxes", "scores" and "masks"
            img_shape (Tuple[int, int]): height and width of original image

        Returns:
            boxes (np.ndarray): postprocessed array of detected bboxes
            scores (np.ndarray): postprocessed array of scores
            labels (np.ndarray): postprocessed array of labels
            masks (np.ndarray): postprocessed array of binarized masks
        """
        masks = np.empty((0, 0, 0), dtype=np.uint8)
        scores = np.empty((0), dtype=np.float32)
        labels = np.empty((0))
        bboxes = np.empty((0, 4), dtype=np.float32)
        if len(network_output["labels"]) > 0:
            network_output["labels"] -= 1

            # Indices to filter out unwanted classes
            detect_filter = torch.where(
                torch.isin(network_output["labels"], self.detect_ids)
            )

            for output_key in network_output.keys():
                self.filtered_output[output_key] = network_output[output_key][
                    detect_filter
                ]

            if len(self.filtered_output["labels"]) > 0:
                bboxes = self.filtered_output["boxes"].cpu().numpy()
                # Normalize the bbox coordinates by the image size
                bboxes = xyxy2xyxyn(bboxes, height=img_shape[0], width=img_shape[1])
                bboxes = np.clip(bboxes, 0, 1)

                label_numbers = self.filtered_output["labels"].cpu().numpy()
                labels = np.vectorize(self.class_names.get)(label_numbers)

                scores = self.filtered_output["scores"].cpu().numpy()

                # Binarize mask's pixel values by confidence score
                masks = self.filtered_output["masks"] > self.mask_threshold
                masks = masks.squeeze(1).cpu().numpy().astype(np.uint8)

        return bboxes, labels, scores, masks

    def _preprocess(self, image: np.ndarray) -> List[np.ndarray]:
        """Preprocessing function for mask_rcnn

        Args:
            image (np.ndarray): Image in numpy array.

        Returns:
            image (np.ndarray): The preprocessed image
        """
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return [Detector.preprocess_transform(image_rgb).to(self.device)]



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Region of Interest Heads for Faster R-CNN.
Modifications include:
- Removed training, target and losses related code
- Removed keypoint detection related codes and arguments.
- Removed ONNX related code
"""

from typing import Dict, List, Optional, Tuple, Union
import torch
from torch import nn, Tensor
import torch.nn.functional as F
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection import (
    detection_utils as det_utils,
)
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.ops import (
    boxes as box_ops,
)


def maskrcnn_inference(mask_logits: Tensor, labels: List[Tensor]) -> List[Tensor]:
    """
    From the results of the CNN, post process the masks
    by taking the mask corresponding to the class with max
    probability (which are of fixed size and directly output
    by the CNN) and return the masks in the mask field of the BoxList.

    Args:
        mask_logits (Tensor): The mask logits
        labels (list[BoxList]): bounding boxes that are used as
            reference, one for each image

    Returns:
        results (list[BoxList]): one BoxList for each image, containing
            the extra field mask
    """
    mask_prob = mask_logits.sigmoid()

    # select masks corresponding to the predicted classes
    num_masks = mask_logits.shape[0]
    boxes_per_image = [label.shape[0] for label in labels]
    labels_ = torch.cat(labels)
    index = torch.arange(num_masks, device=labels_.device)
    mask_prob = mask_prob[index, labels_][:, None]
    mask_prob_out = mask_prob.split(boxes_per_image, dim=0)

    return mask_prob_out


def expand_boxes(boxes: Tensor, scale: float) -> Tensor:
    """Enlarge boxes with the specified scale"""
    w_half = (boxes[:, 2] - boxes[:, 0]) * 0.5
    h_half = (boxes[:, 3] - boxes[:, 1]) * 0.5
    x_c = (boxes[:, 2] + boxes[:, 0]) * 0.5
    y_c = (boxes[:, 3] + boxes[:, 1]) * 0.5

    w_half *= scale
    h_half *= scale

    boxes_exp = torch.zeros_like(boxes)
    boxes_exp[:, 0] = x_c - w_half
    boxes_exp[:, 2] = x_c + w_half
    boxes_exp[:, 1] = y_c - h_half
    boxes_exp[:, 3] = y_c + h_half
    return boxes_exp


def expand_masks(mask: Tensor, padding: int) -> Tuple[Tensor, float]:
    """Enlarge masks with the specified padding and returns the resultant scale"""
    mask_width = mask.shape[-1]
    scale = float(mask_width + 2 * padding) / mask_width
    padded_mask = F.pad(mask, (padding,) * 4)
    return padded_mask, scale


def paste_mask_in_image(mask: Tensor, box: Tensor, im_h: int, im_w: int) -> Tensor:
    """Resize mask to same size as the bounding box and paste onto an image-size frame that is
    initialized to zero. The location of the mask on the frame follows the location of the bounding
    box"""
    to_remove = 1
    width = int(box[2] - box[0] + to_remove)
    height = int(box[3] - box[1] + to_remove)
    width = max(width, 1)
    height = max(height, 1)

    # Set shape to [batchxCxHxW]
    mask = mask.expand((1, 1, -1, -1))

    # Resize mask
    mask = F.interpolate(
        mask, size=(height, width), mode="bilinear", align_corners=False
    )
    mask = mask[0][0]

    im_mask = torch.zeros((im_h, im_w), dtype=mask.dtype, device=mask.device)
    x_0 = max(box[0], 0)  # type: ignore[call-overload]
    x_1 = min(box[2] + 1, im_w)  # type: ignore[call-overload]
    y_0 = max(box[1], 0)  # type: ignore[call-overload]
    y_1 = min(box[3] + 1, im_h)  # type: ignore[call-overload]

    im_mask[y_0:y_1, x_0:x_1] = mask[
        (y_0 - box[1]) : (y_1 - box[1]), (x_0 - box[0]) : (x_1 - box[0])
    ]
    return im_mask


def paste_masks_in_image(
    masks: Tensor, boxes: Tensor, img_shape: Tuple[int, int], padding: int = 1
) -> Tensor:
    """Expands masks and bounding boxes based on the padding size and paste masks onto a
    zero-initialized frame and a location based on the bounding boxes using the
    `paste_mask_in_image()` function"""
    masks, scale = expand_masks(masks, padding=padding)
    boxes = expand_boxes(boxes, scale).to(dtype=torch.int64)
    im_h, im_w = img_shape

    res = [paste_mask_in_image(m[0], b, im_h, im_w) for m, b in zip(masks, boxes)]
    if len(res) > 0:
        ret = torch.stack(res, dim=0)[:, None]
    else:
        ret = masks.new_empty((0, 1, im_h, im_w))
    return ret


class RoIHeads(nn.Module):
    """A class for Region of Interest Head for Mask-RCNN

    Args:
        box_roi_pool (nn.Module): (MultiScaleRoIAlign): the module which crops and resizes the
            feature maps in the locations indicated by the bounding boxes
        box_head (nn.Module): module that takes the cropped feature maps as input
        box_predictor (nn.Module): module that takes the output of box_head and returns the
            classification logits and box regression deltas.
        bbox_reg_weights (Tuple[float, float, float, float]): weights for the
            encoding/decoding of the bounding boxes
        score_thresh (float): during inference, only return proposals with a classification
            score greater than box_score_thresh
        nms_thresh (float): NMS threshold for the prediction head.
        detections_per_img (int): maximum number of detections per image, for all classes.
        mask_roi_pool (nn.Module, optional): the module which crops and resizes the
            feature maps in the locations indicated by the bounding boxes, which will be used for
            the mask head. Defaults to None.
        mask_head (nn.Module, optional): module that takes the cropped feature maps as
            input. Defaults to None.
        mask_predictor (nn.Module, optional): module that takes the output of the
            mask_head and returns the segmentation mask logits. Defaults to None.
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(
        self,
        box_roi_pool: nn.Module,
        box_head: nn.Module,
        box_predictor: nn.Module,
        bbox_reg_weights: Union[Tuple[float, float, float, float], None],
        # Faster R-CNN inference
        score_thresh: float,
        nms_thresh: float,
        detections_per_img: int,
        # Mask
        mask_roi_pool: Optional[nn.Module] = None,
        mask_head: Optional[nn.Module] = None,
        mask_predictor: Optional[nn.Module] = None,
    ):
        super().__init__()

        if bbox_reg_weights is None:
            bbox_reg_weights = (10.0, 10.0, 5.0, 5.0)
        self.box_coder = det_utils.BoxCoder(bbox_reg_weights)

        self.box_roi_pool = box_roi_pool
        self.box_head = box_head
        self.box_predictor = box_predictor

        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img

        self.mask_roi_pool = mask_roi_pool
        self.mask_head = mask_head
        self.mask_predictor = mask_predictor

    def has_mask(self) -> bool:
        """Checks whether the RoI heads have mask_roi_pool, mask_head and mask_predictor

        Returns:
            bool: True if the RoI heads have mask_roi_pool, mask_head and mask_predictor
        """
        if self.mask_roi_pool is None:
            return False
        if self.mask_head is None:
            return False
        if self.mask_predictor is None:
            return False
        return True

    def postprocess_detections(
        self,
        class_logits: Tensor,
        box_regression: Tensor,
        proposals: List[Tensor],
        image_shapes: List[Tuple[int, int]],
    ) -> Tuple[List[Tensor], List[Tensor], List[Tensor]]:
        # pylint: disable=too-many-locals
        """Perform postprocessing for detection results"""
        device = class_logits.device
        num_classes = class_logits.shape[-1]

        boxes_per_image = [boxes_in_image.shape[0] for boxes_in_image in proposals]
        pred_boxes = self.box_coder.decode(box_regression, proposals)

        pred_scores = F.softmax(class_logits, -1)

        pred_boxes_list = pred_boxes.split(boxes_per_image, 0)
        pred_scores_list = pred_scores.split(boxes_per_image, 0)

        all_boxes = []
        all_scores = []
        all_labels = []
        for boxes, scores, image_shape in zip(
            pred_boxes_list, pred_scores_list, image_shapes
        ):
            boxes = box_ops.clip_boxes_to_image(boxes, image_shape)

            # create labels for each prediction
            labels = torch.arange(num_classes, device=device)
            labels = labels.view(1, -1).expand_as(scores)

            # remove predictions with the background label
            boxes = boxes[:, 1:]
            scores = scores[:, 1:]
            labels = labels[:, 1:]

            # batch everything, by making every class prediction be a separate instance
            boxes = boxes.reshape(-1, 4)
            scores = scores.reshape(-1)
            labels = labels.reshape(-1)

            # remove low scoring boxes
            inds = torch.where(scores > self.score_thresh)[0]
            boxes, scores, labels = boxes[inds], scores[inds], labels[inds]

            # remove empty boxes
            keep = box_ops.remove_small_boxes(boxes, min_size=1e-2)
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            # non-maximum suppression, independently done per class
            keep = box_ops.batched_nms(boxes, scores, labels, self.nms_thresh)
            # keep only topk scoring predictions
            keep = keep[: self.detections_per_img]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            all_boxes.append(boxes)
            all_scores.append(scores)
            all_labels.append(labels)

        return all_boxes, all_scores, all_labels

    def forward(
        self,
        features: Dict[str, Tensor],
        proposals: List[Tensor],
        image_shapes: List[Tuple[int, int]],
    ) -> List[Dict[str, Tensor]]:
        # pylint: disable=too-many-locals
        """Forward propagation for the RoIHead.
        Takes the feature and proposals, and predicts the bounding boxes, labels, scores and
        predicts the masks if there are any mask_roi_pool, mask_head and mask_predictor present.

        Args:
            features (Dict[str, Tensor]): Features from the backbone
            proposals (List[Tensor]): Proposals from the RPN
            image_shapes (List[Tuple[int, int]]): Original sizes of the input images

        Raises:
            Exception: If mask_roi_pool is not present even after has_mask() return True

        Returns:
            List[Dict[str, Tensor]]: A list of dictionary of prediction outputs. The keys in the
                dictionary are:
                - boxes (FloatTensor[N, 4]): the predicted boxes in [x1, y1, x2, y2] format, with
                    0 <= x1 < x2 <= W and 0 <= y1 < y2 <= H.
                - labels (Int64Tensor[N]): the predicted labels for each image
                - scores (Tensor[N]): the scores of each prediction
                - masks (FloatTensor[N, 1, H, W]): the predicted masks for each instance with
                    probability values [0, 1]
        """
        labels = None

        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        class_logits, box_regression = self.box_predictor(box_features)

        results: List[Dict[str, torch.Tensor]] = []
        boxes, scores, labels = self.postprocess_detections(
            class_logits, box_regression, proposals, image_shapes
        )
        num_images = len(boxes)
        for i in range(num_images):
            results.append(
                {
                    "boxes": boxes[i],
                    "labels": labels[i],
                    "scores": scores[i],
                }
            )

        if self.has_mask():
            mask_proposals = [result["boxes"] for result in results]

            if self.mask_roi_pool is not None:
                mask_features = self.mask_roi_pool(
                    features, mask_proposals, image_shapes
                )
                mask_features = self.mask_head(mask_features)  # type: ignore[misc]
                mask_logits = self.mask_predictor(mask_features)  # type: ignore[misc]
            else:
                raise Exception("Expected mask_roi_pool to be not None")

            labels = [result["labels"] for result in results]
            masks_probs = maskrcnn_inference(mask_logits, labels)
            for mask_prob, result in zip(masks_probs, results):
                result["masks"] = mask_prob

        return results



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws instance segmentation masks.
"""

import colorsys
from random import randint
from pydoc import locate
from typing import Any, Callable, Dict, List, Tuple, cast

import cv2
import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.base import ThresholdCheckerMixin
from peekingduck.pipeline.nodes.draw.utils.constants import (
    SATURATION_STEPS,
    SATURATION_MINIMUM,
    ALPHA,
    CONTOUR_COLOR,
    CLASS_COLORS,
    DEFAULT_CLASS_COLOR,
)


class Node(AbstractNode, ThresholdCheckerMixin):
    # pylint: disable=line-too-long
    """Draws instance segmentation masks on image.

    The :mod:`draw.mask` node draws instance segmentation masks onto the
    detected object instances.

    Inputs:
        |img_data|

        |masks_data|

        |bbox_labels_data|

    Outputs:
        |img_data|

    Configs:
        instance_color_scheme (:obj:`str`): **{"random", "hue_family"},
            default = "hue_family"** |br|
            This defines what colors to use for the standard masks. |br|
            "hue_family": use the same hue for each instance belonging to
            the same class, but with a slightly different saturation. |br|
            "random": use a random color for all instances.
        effect (:obj:`dict`): **{contrast: null, brightness: null,
            gamma_correction: null, blur: null, mosaic: null}** |br|
            This defines the effect (if any) to apply to either the masked
            (objects) or unmasked (background) areas of the image. If no effect
            is selected, a "standard" instance segmentation mask will be drawn
            and colored according to the instance_color_scheme. |br|
            For example, to apply the contrast effect to the objects in the
            image, set the following config in pipeline_config.yml:

                effect : {contrast: 1.2}

            Note that at most one effect can be enabled at a time.

            +------------------+---------------------------------------+-----------+-------------+
            | Effect           | Description                           | Data Type | Range       |
            +------------------+---------------------------------------+-----------+-------------+
            | contrast         | Adjusts contrast using this value     | float     | [0.0, 3.0]  |
            |                  | as the "alpha" parameter.             |           |             |
            +------------------+---------------------------------------+-----------+-------------+
            | brightness       | Adjusts brightness using this         | int       | [-100, 100] |
            |                  | value as the "beta" parameter.        |           |             |
            +------------------+---------------------------------------+-----------+-------------+
            | gamma_correction | Adjusts gamma using this value as     | float     | [0.0, +inf] |
            |                  | the "gamma" parameter.                |           |             |
            +------------------+---------------------------------------+-----------+-------------+
            | blur             | Blurs the area using this value as    | int       | [1, +inf]   |
            |                  | the "blur_kernel_size" parameter. |br||           |             |
            |                  | Larger values gives more intense      |           |             |
            |                  | blurring.                             |           |             |
            +------------------+---------------------------------------+-----------+-------------+
            | mosaic           | Mosaics the area using this value     | int       | [1, +inf]   |
            |                  | as the resolution of a mosaic         |           |             |
            |                  | filter |br| (width |times| height).   |           |             |
            |                  | The number corresponds to the number  |           |             |
            |                  | of rows |br| and columns used to      |           |             |
            |                  | create a mosaic. For example, the     |           |             |
            |                  | setting |br| (``mosaic: 25``) creates |           |             |
            |                  | a :math:`25 \\times 25` mosaic         |           |             |
            |                  | filter. Increasing |br| the number    |           |             |
            |                  | increases the intensity of            |           |             |
            |                  | pixelation over an area.              |           |             |
            +------------------+---------------------------------------+-----------+-------------+

        effect_area (:obj:`str`): **{"objects", "background"}, default =
            "objects"** |br|
            This defines where the effect should be applied. |br|
            "objects": the effect is applied to the masked areas of the
            image. |br|
            "background": the effect is applied to the unmasked areas of the
            image.
        contours (:obj:`dict`): **{show: False, thickness: 2}** |br|

            +-----------+------------------------------------------------------------+-----------+-----------+
            | Contours  | Description                                                | Data Type | Range     |
            +-----------+------------------------------------------------------------+-----------+-----------+
            | show      | This determines whether to show the contours of the masks. | bool      | N.A.      |
            +-----------+------------------------------------------------------------+-----------+-----------+
            | thickness | This defines the thickness of the contours.                | int       | [1, +inf] |
            +-----------+------------------------------------------------------------+-----------+-----------+
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.class_instance_colors: Dict[str, List[Tuple[int, int, int]]] = {}
        self._validate_configs()

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads an input image and returns the image (1) with the instance
        segmentation masks drawn over the detected object instances, or
        (2) with a visual effect mask applied to it. The visual effect can be
        applied to the masked areas of the image or to the unmasked areas.

        Args:
            inputs (dict): Dictionary of inputs with keys "img", "masks",
            "bbox_labels".

        Returns:
            outputs (dict): Output in dictionary format with key "img".
        """
        if self.effect is None:
            output_img = self._draw_standard_masks(
                inputs["img"],
                inputs["masks"],
                inputs["bbox_labels"],
            )
        else:
            output_img = self._mask_apply_effect(
                inputs["img"],
                inputs["masks"],
                self.effect,
            )

        return {"img": output_img}

    def _validate_configs(self) -> None:
        self.check_valid_choice("instance_color_scheme", {"random", "hue_family"})

        effects_count = 0
        self.effect = None
        for effect, setting in self.config["effect"].items():
            if setting:
                effects_count += 1
                self.effect = effect

        if effects_count > 1:
            raise ValueError("A maximum of one effect can be enabled at a time.")

        self.check_valid_choice("effect_area", {"objects", "background"})

        valid_types = [
            "effect|contrast, float",
            "effect|brightness, int",
            "effect|gamma_correction, float",
            "effect|blur, int",
            "effect|mosaic, int",
            "contours|show, bool",
            "contours|thickness, int",
        ]

        valid_ranges = [
            "effect|contrast, [0.0, 3.0]",
            "effect|brightness, [-100, 100]",
            "effect|gamma_correction, [0.0, +inf]",
            "effect|blur, [1, +inf]",
            "effect|mosaic, [1, +inf]",
            "contours|thickness, [1, +inf]",
        ]

        self._check_configs(valid_types, criteria="type")
        self._check_configs(valid_ranges, criteria="range")

    def _check_configs(self, valid_settings: List[str], criteria: str) -> None:
        if criteria == "type":
            self._check_config_values(valid_settings, Node._type_checker)
        elif criteria == "range":
            self._check_config_values(valid_settings, Node._range_checker)
        else:
            raise ValueError("'criteria' parameter must be either 'type' or 'range'.")

    @staticmethod
    def _type_checker(var_name: str, value: Any, var_type: str) -> None:
        """Checks if the value is of the correct type."""
        if value is not None and not isinstance(value, locate(var_type)):  # type: ignore
            raise ValueError(f"{var_name} must be a {var_type} value.")

    @staticmethod
    def _range_checker(var_name: str, value: Any, number_range: str) -> None:
        """Checks if a value is within a range.  The format for the number
        range is "[lower, upper]"."""
        if value is not None:
            lower, upper = [
                float(value.strip()) for value in number_range[1:-1].split(",")
            ]
            if not lower <= value <= upper:
                raise ValueError(
                    f"{var_name} must be within the range of {number_range}."
                )

    def _check_config_values(
        self, valid_settings: List[str], checker: Callable
    ) -> None:
        """Checks the configs for valid data. The checker parameter determines
        the criteria that the config values are checked against. The
        valid_settings can handle nested configs, with names of different
        levels separated by a pipe (|).

        Example to illustrate format of 'valid_settings' parameter:

            config = {
                "not_nested": 10,
                "single_level_dict": {
                    "contrast": 50.0,
                    "text_setting": "sample string",
                },
                "nested_dict": {
                    "effect": {
                        "blur": 50,
                    }
                },
            }

            valid_settings_for_types = [
                "not_nested, int",
                "single_level_dict|contrast, float",
                "single_level_dict|text_setting, str",
                "nested_dict|effect|blur, int",
            ]

            valid_settings_for_ranges = [
                "not_nested, [1, 10]",
                "single_level_dict|contrast, [0, 3]",
                "nested_dict|effect|blur, [1, +inf]",
            ]
        """
        for valid_setting in valid_settings:
            criteria_input_list = valid_setting.split(",", maxsplit=1)
            var_location, var_criteria = (
                criteria_input_list[0],
                criteria_input_list[1].strip(),
            )
            variables = var_location.split("|")

            var_name = "config"
            dict_name = self.config
            while len(variables) > 1:
                var_name += f'["{variables[0]}"]'
                dict_name = dict_name.get(variables.pop(0))  # type: ignore

            var_name += f'["{variables[0]}"]'

            checker(var_name, dict_name[variables[0]], var_criteria)

    def _draw_standard_masks(  # pylint: disable-msg=too-many-locals
        self,
        image: np.ndarray,
        masks: np.ndarray,
        bbox_labels: np.ndarray,
    ) -> np.ndarray:
        """Draws instance segmentation masks over detected objects.

        Args:
            image (numpy.ndarray): Input image.
            masks (numpy.ndarray): NumPy array of binary (0/1) masks, one mask
                for each detected object, in the same order as bbox_labels.
            bbox_labels (numpy.ndarray): NumPy array of strings representing
                the labels of detected objects. The order corresponds to
                ``masks``.

        Returns:
            numpy.ndarray: Input image with instance segmentation masks
                applied to it.
        """
        self.class_instance_counts: Dict[str, int] = {}

        full_sized_canvas = np.zeros(image.shape, image.dtype)
        ret_image = image

        for index, _ in enumerate(bbox_labels):
            color = self._get_instance_color(bbox_labels[index])

            full_sized_canvas[:, :] = color

            coloured_seg_mask = cv2.bitwise_and(
                full_sized_canvas, full_sized_canvas, mask=masks[index]
            )
            masked_area = cv2.bitwise_and(image, image, mask=masks[index])
            masked_area_colored = cv2.addWeighted(
                coloured_seg_mask, ALPHA, masked_area, 1 - ALPHA, 0
            )

            # get the inverted mask i.e. image outside of the masked area
            mask_inv = 1 - masks[index]
            # remove masked area from image to be returned
            ret_image = cv2.bitwise_and(ret_image, ret_image, mask=mask_inv)
            ret_image = cv2.add(ret_image, masked_area_colored)

            if self.config["contours"]["show"]:
                ret_image = self._draw_contours(masks, ret_image, index)

        return ret_image

    def _get_instance_color(self, instance_class: str) -> Tuple[int, int, int]:
        """Returns color to use for next segmentation instance according to the
        chosen instance color scheme.

        Args:
            instance_class (str): Class of detected object.

        Returns:
            Tuple[int,int,int]: Color to use for next instance, in BGR.
        """
        self.class_instance_counts[instance_class] = (
            self.class_instance_counts.get(instance_class, 0) + 1
        )

        if self.class_instance_counts[instance_class] > len(
            self.class_instance_colors.setdefault(instance_class, [])
        ):
            # get new color
            color = self._get_new_instance_color(instance_class)
            # append new assigned color to instance_colors
            self.class_instance_colors[instance_class].append(color)
        else:
            # get color already assigned to this instance number
            color = self.class_instance_colors[instance_class][
                self.class_instance_counts[instance_class] - 1
            ]

        return color

    def _get_new_instance_color(self, instance_class: str) -> Tuple[int, int, int]:
        """Returns color to use for a new segmentation instance. When we
        encounter an instance that is one more than the count of instances of
        a particular class, the instance has not been assigned a color yet
        (from previous frames) and we need to get a new color.

        Depending on the chosen instance color scheme, this is how we determine
        the color:

        "random" - Random color.

        "hue_family" - Each object class uses a certain pre-assigned hue. Then
            each new instance will use the same hue, but with a slightly
            different saturation.

        Args:
            instance_class (str): Class of detected object.

        Returns:
            Tuple[int,int,int]: Color to use for next instance, in BGR.
        """
        if self.config["instance_color_scheme"] == "random":
            color_hsv = (randint(0, 179), randint(100, 255), 255)
            color = self._hsv_to_bgr(color_hsv)
        elif self.config["instance_color_scheme"] == "hue_family":
            if not self.class_instance_colors.get(instance_class):
                color = CLASS_COLORS.get(instance_class, DEFAULT_CLASS_COLOR)
            else:
                color = self.class_instance_colors.get(instance_class)[-1]  # type: ignore
                color_hsv = self._bgr_to_hsv(color)
                # we use a minimum saturation of 100 to avoid too light colors,
                # thus we increment saturation by step size of (256-100)/8.
                saturation = (
                    color_hsv[1] + (256 - SATURATION_MINIMUM) / SATURATION_STEPS
                ) % 256
                if saturation < SATURATION_MINIMUM:
                    saturation += SATURATION_MINIMUM
                color_hsv = (color_hsv[0], int(saturation), color_hsv[2])
                color = self._hsv_to_bgr(color_hsv)

        return color

    def _draw_contours(
        self,
        masks: np.ndarray,
        image: np.ndarray,
        index: int = None,
    ) -> np.ndarray:
        """Draws contours around instance segmentation masks. If 'index' is
        given, only the contour of the mask with the given index is drawn."""
        ret_image = image
        masks_to_process = [index] if index else range(masks.shape[0])
        for i in masks_to_process:
            contour, _ = cv2.findContours(
                masks[i], cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
            )
            cv2.drawContours(
                ret_image,
                contour,
                -1,
                CONTOUR_COLOR,
                self.config["contours"]["thickness"],
            )

        return ret_image

    def _mask_apply_effect(
        self, image: np.ndarray, masks: np.ndarray, effect: str
    ) -> np.ndarray:
        """Applies the chosen effect to the image and masks."""
        combined_masks = np.zeros(image.shape[:2], dtype="uint8")
        # combine all the individual masks
        for mask in masks:
            np.putmask(combined_masks, mask, 1)

        if self.config["effect_area"] == "objects":
            effect_area = combined_masks
        else:  # effect_area == "background"
            effect_area = 1 - combined_masks

        if effect in ["contrast", "brightness"]:
            full_image_with_effect = self._adjust_contrast_brightness(image, effect)
        elif effect == "gamma_correction":
            full_image_with_effect = self._gamma_correction(image)
        elif effect == "blur":
            full_image_with_effect = cv2.blur(
                image,
                (self.config["effect"]["blur"], self.config["effect"]["blur"]),
            )
        elif effect == "mosaic":
            full_image_with_effect = self._mosaic_image(image)

        effect_area_with_effect = cv2.bitwise_and(
            full_image_with_effect, full_image_with_effect, mask=effect_area
        )
        # get the inverted mask
        effect_area_inv = 1 - effect_area
        # remove effect area from original image
        image_empty_effect_area = cv2.bitwise_and(image, image, mask=effect_area_inv)
        # add both images together
        ret_image = cv2.add(image_empty_effect_area, effect_area_with_effect)

        if self.config["contours"]["show"]:
            ret_image = self._draw_contours(masks, ret_image)

        return ret_image

    def _adjust_contrast_brightness(self, image: np.ndarray, effect: str) -> np.ndarray:
        if effect == "contrast":
            alpha = self.config["effect"]["contrast"]
            beta = 0
        else:  # i.e. effect == "brightness"
            alpha = 1
            beta = self.config["effect"]["brightness"]

        return cv2.convertScaleAbs(image, alpha=alpha, beta=beta)

    def _gamma_correction(self, image: np.ndarray) -> np.ndarray:
        lookup_table = np.empty((1, 256), np.uint8)
        for i in range(256):
            lookup_table[0, i] = np.clip(
                pow(i / 255.0, self.config["effect"]["gamma_correction"]) * 255.0,
                0,
                255,
            )

        image_gamma_corrected = cv2.LUT(image, lookup_table)

        return image_gamma_corrected

    def _mosaic_image(self, image: np.ndarray) -> np.ndarray:
        height, width = image.shape[:2]
        ret_image = cv2.resize(
            image,
            (self.config["effect"]["mosaic"], self.config["effect"]["mosaic"]),
            interpolation=cv2.INTER_LANCZOS4,
        )
        ret_image = cv2.resize(
            ret_image, (width, height), interpolation=cv2.INTER_NEAREST
        )

        return ret_image

    @staticmethod
    def _bgr_to_hsv(bgr: Tuple[int, int, int]) -> Tuple[int, int, int]:
        hsv_norm = colorsys.rgb_to_hsv(bgr[2] / 255, bgr[1] / 255, bgr[0] / 255)
        hsv = (int(hsv_norm[0] * 127), int(hsv_norm[1] * 255), int(hsv_norm[2] * 255))

        return hsv

    @staticmethod
    def _hsv_to_bgr(hsv: Tuple[int, int, int]) -> Tuple[int, int, int]:
        rgb_norm = colorsys.hsv_to_rgb(hsv[0] / 127, hsv[1] / 255, hsv[2] / 255)
        rgb = [int(x * 255) for x in rgb_norm]
        bgr = cast(Tuple[int, int, int], tuple(rgb[::-1]))

        return bgr



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""An object that holds a list of padded images along with its original sizes"""

from typing import List, Tuple
import torch
from torch import Tensor


class ImageList:
    """
    Structure that holds a list of images (of possibly
    varying sizes) as a single tensor.
    This works by padding the images to the same size,
    and storing in a field the original sizes of each image
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, tensors: Tensor, image_sizes: List[Tuple[int, int]]):
        """
        Args:
            tensors (Tensor)
            image_sizes (list[tuple[int, int]])
        """
        self.tensors = tensors
        self.image_sizes = image_sizes

    def to(self, device: torch.device) -> "ImageList":  # pylint: disable=invalid-name
        """Transfer the images to target device"""
        cast_tensor = self.tensors.to(device)
        return ImageList(cast_tensor, self.image_sizes)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mask R-CNN model with backbone type: r50-fpn and r101-fpn
"""

import logging
from typing import Any, Dict, List, Tuple
import json
import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detector import (
    Detector,
)


class MaskRCNNModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """Mask R-CNN model with ResNet 50 / 101 FPN backbone"""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds(
            ["iou_threshold", "score_threshold", "mask_threshold"], "[0, 1]"
        )
        self.check_bounds(["min_size", "max_size", "max_num_detections"], "[1 , +inf)")

        model_dir = self.download_weights()
        classes_path = model_dir / self.weights["classes_file"]
        class_names = {
            val["id"] - 1: val["name"]
            for val in json.loads(classes_path.read_text()).values()
        }

        self.detect_ids = self.config["detect"]
        self.detector = Detector(
            model_dir,
            class_names,
            self.detect_ids,
            self.config["model_type"],
            self.config["num_classes"],
            self.weights["model_file"],
            self.config["min_size"],
            self.config["max_size"],
            self.config["iou_threshold"],
            self.config["max_num_detections"],
            self.config["score_threshold"],
            self.config["mask_threshold"],
        )

    @property
    def detect_ids(self) -> List[int]:
        """The list of selected object category IDs."""
        return self._detect_ids

    @detect_ids.setter
    def detect_ids(self, ids: List[int]) -> None:
        if not isinstance(ids, list):
            raise TypeError("detect_ids has to be a list")
        if not ids:
            self.logger.info("Detecting all Mask R-CNN classes")
        self._detect_ids = ids

    def predict(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Predicts bboxes and masks from image.

        Args:
            image (np.ndarray): Input image frame.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Returned tuple
            contains:
            - An array of detection bboxes
            - An array of human-friendly detection class names
            - An array of detection scores
            - An array of binarized masks

        Raises:
            TypeError: The provided `image` is not a numpy array.
        """
        if not isinstance(image, np.ndarray):
            raise TypeError("image must be a np.ndarray")
        return self.detector.predict_instance_mask_from_image(image)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🎭 Instance segmentation model for real-time inference"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.model.yolact_edgev1 import yolact_edge_model
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Initializes and uses YolactEdge to infer from an image frame

    The YolactEdge node is capable of detecting objects from 80 categories.
    The table of object categories can be found
    :ref:`here <general-object-detection-ids>`.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_labels_data|

        |bbox_scores_data|

        |masks_data|

    Configs:
        model_type (:obj:`str`): (:obj:`str`): **{"r101-fpn", "r50-fpn",
            "mobilenetv2"}, default="r50-fpn"**. |br|
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        input_size (:obj:`int`): **default = 550**. |br|
            Input image resolution of the YolactEdge model.
        detect (:obj:`List[Union[int, string]]`): **default=[0]**. |br|
            List of object class names or IDs to be detected. To detect all classes,
            refer to the :ref:`tech note <general-object-detection-ids>`.
        max_num_detections: (:obj:`int`): **default=100**. |br|
            Maximum number of detections per image, for all classes.
        iou_threshold (:obj:`float`): **[0, 1], default = 0.5**. |br|
            Overlapping bounding boxes with Intersection over Union (IoU) above
            the threshold will be discarded.
        score_threshold (:obj:`float`): **[0, 1], default = 0.2**. |br|
            Bounding boxes with confidence score (product of objectness score
            and classification score) below the threshold will be discarded.


    References:
        YolactEdge: Real-time Instance Segmentation on the Edge
        https://arxiv.org/abs/2012.12259

        Inference code and model weights:
        https://github.com/haotian-liu/yolact_edge
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = yolact_edge_model.YolactEdgeModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads `img` from `inputs` and return the bboxes and masks of the detected
        objects.

        The classes of objects to be detected can be specified through the
        `detect` configuration option.

        Args:
            inputs (Dict): Inputs dictionary with the key `img`.

        Returns:
            (Dict): Outputs dictionary with the keys `bboxes`, `bbox_labels`,
                `bbox_scores` and `masks`.
        """
        bboxes, labels, scores, masks = self.model.predict(inputs["img"])

        outputs = {
            "bboxes": bboxes,
            "bbox_labels": labels,
            "bbox_scores": scores,
            "masks": masks,
        }

        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🎭 Instance segmentation model for generating high-quality masks."""

from typing import Any, Dict

from peekingduck.pipeline.nodes.model.mask_rcnnv1 import mask_rcnn_model
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Initializes and uses Mask R-CNN to infer from an image frame.

    The Mask-RCNN node is capable detecting objects and their respective masks
    from 80 categories. The table of object categories can be found
    :ref:`here <general-instance-segmentation-ids>`. The ``"r50-fpn"`` backbone is
    used by default, and the ``"r101-fpn"`` for the ResNet 101 backbone variant can also
    be chosen.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_labels_data|

        |bbox_scores_data|

        |masks_data|

    Configs:
        model_type (:obj:`str`): **{"r50-fpn", "r101-fpn"}, default = "r50-fpn"**. |br|
            Defines the type of backbones to be used.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        min_size (:obj:`int`): **default = 800**. |br|
            Minimum size of the image to be rescaled before feeding it to the
            backbone.
        max_size (:obj:`int`): **default = 1333**. |br|
            Maximum size of the image to be rescaled before feeding it to the
            backbone.
        detect (:obj:`List[Union[int, string]]`): **default = [0]**. |br|
            List of object class names or IDs to be detected. To detect all classes,
            refer to the :ref:`tech note <general-instance-segmentation-ids>`.
        max_num_detections: (:obj:`int`): **default = 100**. |br|
            Maximum number of detections per image, for all classes.
        iou_threshold (:obj:`float`): **[0, 1], default = 0.5**. |br|
            Overlapping bounding boxes with Intersection over Union (IoU) above
            the threshold will be discarded.
        score_threshold (:obj:`float`): **[0, 1], default = 0.5**. |br|
            Bounding boxes with classification score below the threshold will be discarded.
        mask_threshold (:obj:`float`): **[0, 1], default = 0.5**. |br|
            The confidence threshold for binarizing the masks' pixel values; determines whether an
            object is detected at a particular pixel.

    References:
        Mask R-CNN: A conceptually simple, flexible, and general framework for object
        instance segmentation.:
        https://arxiv.org/abs/1703.06870

        Inference code adapted from:
        https://pytorch.org/vision/0.11/_modules/torchvision/models/detection/mask_rcnn.html

        The weights for Mask-RCNN Model with ResNet50 FPN backbone were adapted from:
        https://download.pytorch.org/models/maskrcnn_resnet50_fpn_coco-bf2d0c1e.pth
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = mask_rcnn_model.MaskRCNNModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads `img` from `inputs` and return the bboxes and masks of the detect
        objects.

        The classes of objects to be detected can be specified through the
        `detect` configuration option.

        Args:
            inputs (Dict): Inputs dictionary with the key `img`.

        Returns:
            (Dict): Outputs dictionary with the keys `bboxes`, `bbox_labels`,
                `bbox_scores` and `masks`.
        """
        bboxes, labels, scores, masks = self.model.predict(inputs["img"])

        outputs = {
            "bboxes": bboxes,
            "bbox_labels": labels,
            "bbox_scores": scores,
            "masks": masks,
        }

        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Constants used for drawing
"""

# colors
CHAMPAGNE = (156, 223, 244)
BLIZZARD = (241, 232, 164)
VIOLET_BLUE = (188, 118, 119)
TOMATO = (77, 103, 255)
BROWN = (96, 109, 167)
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
PRIMARY_PALETTE_LENGTH = 5
PRIMARY_PALETTE = [CHAMPAGNE, BLIZZARD, TOMATO, VIOLET_BLUE, BROWN]

# constants for thickness
THIN = 1
THICK = 2
VERY_THICK = 3

# constant to fill shapes in cv2. To be replace line thickness
FILLED = -1

# constants for font scale
SMALL_FONTSCALE = 0.5
NORMAL_FONTSCALE = 1
BIG_FONTSCALE = 2

# constants used for image manipulation
LOWER_SATURATION = 0.5

# constants for pts
POINT_RADIUS = 5

# constants for masks
SATURATION_STEPS = 8
SATURATION_MINIMUM = 100
ALPHA = 0.5
CONTOUR_COLOR = (0, 0, 0)
DEFAULT_CLASS_COLOR = (204, 204, 204)
# fmt: off
CLASS_COLORS = {
    "person": CHAMPAGNE,

    "bicycle":      BLIZZARD,
    "car":          BLIZZARD,
    "motorcycle":   BLIZZARD,
    "airplane":     BLIZZARD,
    "bus":          BLIZZARD,
    "train":        BLIZZARD,
    "truck":        BLIZZARD,
    "boat":         BLIZZARD,

    "traffic light":    (79,14,136),
    "fire hydrant":     (79,14,136),
    "stop sign":        (79,14,136),
    "parking meter":    (79,14,136),
    "bench":            (79,14,136),

    "bird":     TOMATO,
    "cat":      TOMATO,
    "dog":      TOMATO,
    "horse":    TOMATO,
    "sheep":    TOMATO,
    "cow":      TOMATO,
    "elephant": TOMATO,
    "bear":     TOMATO,
    "zebra":    TOMATO,
    "giraffe":  TOMATO,

    "backpack": (0,51,0),
    "umbrella": (0,51,0),
    "handbag":  (0,51,0),
    'tie':      (0,51,0),
    "suitcase": (0,51,0),

    "frisbee":          (140,20,74),
    "skis":             (140,20,74),
    "snowboard":        (140,20,74),
    "sports ball":      (140,20,74),
    "kite":             (140,20,74),
    "baseball bat":     (140,20,74),
    "baseball glove":   (140,20,74),
    "skateboard":       (140,20,74),
    "surfboard":        (140,20,74),
    "tennis racket":    (140,20,74),

    "bottle":       (0,111,255),
    "wine glass":   (0,111,255),
    "cup":          (0,111,255),
    "fork":         (0,111,255),
    "knife":        (0,111,255),
    "spoon":        (0,111,255),
    "bowl":         (0,111,255),

    "banana":   BROWN,
    "apple":    BROWN,
    "sandwich": BROWN,
    "orange":   BROWN,
    "broccoli": BROWN,
    "carrot":   BROWN,
    "hot dog":  BROWN,
    "pizza":    BROWN,
    "donut":    BROWN,
    "cake":     BROWN,

    "chair":        (155,87,1),
    "couch":        (155,87,1),
    "potted plant": (155,87,1),
    "bed":          (155,87,1),
    "dining table": (155,87,1),
    "toilet":       (155,87,1),

    "tv":           VIOLET_BLUE,
    "laptop":       VIOLET_BLUE,
    "mouse":        VIOLET_BLUE,
    "remote":       VIOLET_BLUE,
    "keyboard":     VIOLET_BLUE,
    "cell phone":   VIOLET_BLUE,

    "microwave":    (12,54,191),
    "oven":         (12,54,191),
    "toaster":      (12,54,191),
    "sink":         (12,54,191),
    "refrigerator": (12,54,191),

    "book":         (35,39,62),
    "clock":        (35,39,62),
    "vase":         (35,39,62),
    "scissors":     (35,39,62),
    "teddy bear":   (35,39,62),
    "hair drier":   (35,39,62),
    "toothbrush":   (35,39,62),
}
# fmt: on



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Generate FPN for Mask-RCNN backbone"""

from collections import OrderedDict
from typing import Tuple, List, Dict, Optional
import torch.nn.functional as F
from torch import nn, Tensor


class ExtraFPNBlock(nn.Module):
    """Base class for the extra block in the FPN."""

    def forward(
        self,
        results: List[Tensor],
        x_in: List[Tensor],
        names: List[str],
    ) -> Tuple[List[Tensor], List[str]]:
        """A method for forward propagation of the Extra FPN Block. To be overridden by child class
        method.

        Args:
        results (List[Tensor]): the result of the FPN
        x_in (List[Tensor]): the original feature maps
        names (List[str]): the names for each one of the original feature maps

        Returns:
            results (List[Tensor]): the extended set of results of the FPN
            names (List[str]): the extended set of names for the results
        """
        raise NotImplementedError


class FeaturePyramidNetwork(nn.Module):
    """
    Module that adds a FPN from on top of a set of feature maps. This is based on
    `"Feature Pyramid Network for Object Detection" <https://arxiv.org/abs/1612.03144>`_.

    The feature maps are currently supposed to be in increasing depth
    order.

    The input to the model is expected to be an OrderedDict[Tensor], containing
    the feature maps on top of which the FPN will be added.

    Args:
        in_channels_list (list[int]): number of channels for each feature map that
            is passed to the module
        out_channels (int): number of channels of the FPN representation
        extra_blocks (ExtraFPNBlock or None): if provided, extra operations will
            be performed. It is expected to take the fpn features, the original
            features and the names of the original features as input, and returns
            a new list of feature maps and their corresponding names
    """

    def __init__(
        self,
        in_channels_list: List[int],
        out_channels: int,
        extra_blocks: Optional[ExtraFPNBlock] = None,
    ):
        super().__init__()
        self.inner_blocks = nn.ModuleList()
        self.layer_blocks = nn.ModuleList()
        for in_channels in in_channels_list:
            if in_channels == 0:
                raise ValueError("in_channels=0 is currently not supported")
            inner_block_module = nn.Conv2d(in_channels, out_channels, 1)
            layer_block_module = nn.Conv2d(out_channels, out_channels, 3, padding=1)
            self.inner_blocks.append(inner_block_module)
            self.layer_blocks.append(layer_block_module)

        # initialize parameters now to avoid modifying the initialization of top_blocks
        for mod in self.modules():
            if isinstance(mod, nn.Conv2d):
                nn.init.kaiming_uniform_(mod.weight, a=1)
                nn.init.constant_(mod.bias, 0)  # type: ignore[arg-type]

        if extra_blocks is not None:
            assert isinstance(extra_blocks, ExtraFPNBlock)
        self.extra_blocks = extra_blocks

    def get_result_from_inner_blocks(self, x_in: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.inner_blocks[idx](x_in),
        but torchscript doesn't support this yet
        """
        num_blocks = len(self.inner_blocks)
        if idx < 0:
            idx += num_blocks
        i = 0
        out = x_in
        for module in self.inner_blocks:
            if i == idx:
                out = module(x_in)
            i += 1
        return out

    def get_result_from_layer_blocks(self, x_in: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.layer_blocks[idx](x_in),
        but torchscript doesn't support this yet
        """
        num_blocks = len(self.layer_blocks)
        if idx < 0:
            idx += num_blocks
        i = 0
        out = x_in
        for module in self.layer_blocks:
            if i == idx:
                out = module(x_in)
            i += 1
        return out

    def forward(self, x_dict: Dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Computes the FPN for a set of feature maps.

        Args:
            x_dict (OrderedDict[Tensor]): feature maps for each feature level.

        Returns:
            results (OrderedDict[Tensor]): feature maps after FPN layers.
                They are ordered from highest resolution first.
        """
        # unpack OrderedDict into two lists for easier handling
        names = list(x_dict.keys())
        x_list = list(x_dict.values())

        last_inner = self.get_result_from_inner_blocks(x_list[-1], -1)
        results = []
        results.append(self.get_result_from_layer_blocks(last_inner, -1))

        for idx in range(len(x_list) - 2, -1, -1):
            inner_lateral = self.get_result_from_inner_blocks(x_list[idx], idx)
            feat_shape = inner_lateral.shape[-2:]
            inner_top_down = F.interpolate(last_inner, size=feat_shape, mode="nearest")
            last_inner = inner_lateral + inner_top_down
            results.insert(0, self.get_result_from_layer_blocks(last_inner, idx))

        if self.extra_blocks is not None:
            results, names = self.extra_blocks(results, x_list, names)

        # make it back an OrderedDict
        out = OrderedDict((k, v) for k, v in zip(names, results))

        return out


class LastLevelMaxPool(ExtraFPNBlock):
    """
    Applies a max_pool2d on top of the last feature map
    """

    # pylint: disable=unused-argument
    def forward(
        self,
        results: List[Tensor],
        x_in: List[Tensor],
        names: List[str],
    ) -> Tuple[List[Tensor], List[str]]:
        """Performs forward propagation using a 2D max pool layer

        Args:
            results (List[Tensor]): the result of the FPN
            x_in (List[Tensor]): the original feature maps
            names (List[str]): the names for each one of the original feature maps

        Returns:
            results (List[Tensor]): the extended results after 2D max pool of the FPN
            names (List[str]): the extended set of names for the max pooled results
        """
        names.append("pool")
        results.append(F.max_pool2d(results[-1], 1, 2, 0))
        return results, names



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Generate a backbone with FPN for Mask-RCNN model.

Modifications include:
- Removed unused arguments:
    - resnet_fpn_backbone
        - trainable_layers
        - returned_layers
        - extra_blocks
        - norm_layer
"""

from typing import Dict, Optional, List
from collections import OrderedDict
from torch import nn, Tensor
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.ops import (
    misc,
    feature_pyramid_network as fpn,
)
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.backbones import (
    resnet,
)


class BackboneWithFPN(nn.Module):
    """
    Adds a FPN on top of a model.
    Internally, it uses IntermediateLayerGetter to extract a submodel that returns
    the feature maps specified in return_layers.
    The same limitations of IntermediateLayerGetter apply here.
    Args:
        backbone (nn.Module)
        return_layers (Dict[name, new_name]): a dict containing the names
            of the modules for which the activations will be returned as
            the key of the dict, and the value of the dict is the name
            of the returned activation (which the user can specify).
        in_channels_list (List[int]): number of channels for each feature map
            that is returned, in the order they are present in the OrderedDict
        out_channels (int): number of channels in the FPN.
        extra_blocks (ExtraFPNBlock or None): if provided, extra operations will
            be performed. It is expected to take the fpn features, the original
            features and the names of the original features as input, and returns
            a new list of feature maps and their corresponding names. By
            default a ``LastLevelMaxPool`` is used.
    Attributes:
        out_channels (int): the number of channels in the FPN
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        backbone: nn.Module,
        return_layers: Dict[str, str],
        in_channels_list: List[int],
        out_channels: int,
        extra_blocks: Optional[fpn.ExtraFPNBlock] = None,
    ):
        super().__init__()

        if extra_blocks is None:
            extra_blocks = fpn.LastLevelMaxPool()

        self.body = IntermediateLayerGetter(backbone, return_layers=return_layers)
        self.fpn = fpn.FeaturePyramidNetwork(
            in_channels_list=in_channels_list,
            out_channels=out_channels,
            extra_blocks=extra_blocks,
        )
        self.out_channels = out_channels

    def forward(self, inputs: Tensor) -> Dict[str, Tensor]:
        """Takes the input tensor and pass it through the intermediate layers of the model.
        Then, the output from the intermediate layers will be passed to the FPN before returning
        the final output.

        Args:
            inputs (Tensor): Input feature

        Returns:
            Dict[str, Tensor]: An output dictionary. The keys are the names of the returned
                activations of the backbone, and the values are the output features from the FPN
                layers ordered from highest resolution first.
        """
        x_out = self.body(inputs)
        x_out = self.fpn(x_out)
        return x_out


def resnet_fpn_backbone(backbone_name: str) -> BackboneWithFPN:
    """
    Constructs a specified ResNet backbone with FPN on top. Freezes the specified number of
    layers in the backbone.

    Args:
        backbone_name (string): resnet architecture. Possible values are 'resnet50', 'resnet101'

    Returns:
        BackboneWithFPN: An object of the backbone with the FPN attached
    """
    backbone = resnet.__dict__[backbone_name](norm_layer=misc.FrozenBatchNorm2d)

    # select layers that wont be frozen
    for _, parameter in backbone.named_parameters():
        parameter.requires_grad_(False)

    extra_blocks = fpn.LastLevelMaxPool()

    returned_layers = [1, 2, 3, 4]
    return_layers = {f"layer{k}": str(v) for v, k in enumerate(returned_layers)}

    # backbone.inplanes will be equal to the final output feature channels (2048 for both r50
    # and r101), and dividing by 8 will be equal to the number of output channels at the conv2_x
    # layer (layer 1) (256 for both r50 and r101). The elements in the in_channels_list (list of
    # number of input channels for the FPN) should be equal to the number of output channels for
    # the respective layers in the returned_layers, which is increased by a factor of 2 for every
    # subsequent layers (e.g. layer 1 to 4 --> [256, 512, 1024, 2048] for R-50 and R-101).
    in_channels_stage2 = backbone.inplanes // 8
    in_channels_list = [in_channels_stage2 * 2 ** (i - 1) for i in returned_layers]
    out_channels = 256
    return BackboneWithFPN(
        backbone,
        return_layers,
        in_channels_list,
        out_channels,
        extra_blocks=extra_blocks,
    )


class IntermediateLayerGetter(nn.ModuleDict):
    """
    Module wrapper that returns intermediate layers from a model

    It has a strong assumption that the modules have been registered
    into the model in the same order as they are used.
    This means that one should **not** reuse the same nn.Module
    twice in the forward if you want this to work.

    Additionally, it is only able to query submodules that are directly
    assigned to the model. So if `model` is passed, `model.feature1` can
    be returned, but not `model.feature1.layer2`.

    Args:
        model (nn.Module): model on which we will extract the features
        return_layers (Dict[name, new_name]): a dict containing the names
            of the modules for which the activations will be returned as
            the key of the dict, and the value of the dict is the name
            of the returned activation (which the user can specify).
    """

    def __init__(self, model: nn.Module, return_layers: Dict[str, str]) -> None:
        if not set(return_layers).issubset(
            [name for name, _ in model.named_children()]
        ):
            raise ValueError("return_layers are not present in model")
        orig_return_layers = return_layers
        return_layers = {str(k): str(v) for k, v in return_layers.items()}
        layers = OrderedDict()
        for name, module in model.named_children():
            layers[name] = module
            if name in return_layers:
                del return_layers[name]
            if not return_layers:
                break

        super().__init__(layers)
        self.return_layers = orig_return_layers

    def forward(self, inputs: Tensor) -> Dict[str, Tensor]:
        """Takes in a Tensor and obtain the outputs from the respective layers specified in
        self.return_layers.

        Args:
            inputs (Tensor): Input tensor

        Returns:
            Dict[str, Tensor]: A dictionary with keys equal to the enumerated numbers of the
                `return_layers`, and the value is the output features from the respective
                intermediate layers specified in self.return_layers
        """
        out = OrderedDict()
        for name, module in self.items():
            inputs = module(inputs)
            if name in self.return_layers:
                out_name = self.return_layers[name]
                out[out_name] = inputs
        return out
